# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2015, 2016
# --------------------------------------------------------------------------

# pylint:skip-file
# gendoc: ignore

import json
import os

try:
    # python 2
    from urlparse import urlparse
except ImportError:
    # python 3
    from urllib.parse import urlparse


from datetime import datetime
from six import iteritems, string_types
from requests.exceptions import ConnectionError

import warnings

from docplex.util.status import JobSolveStatus

from docloud.job import JobClient, DOcloudInterruptedException, DOcloudNotFoundError
from docloud.status import JobExecutionStatus

from docplex.mp.utils import resolve_pattern, get_logger
from docplex.mp.utils import CyclicLoop

from docplex.mp.context import has_credentials


def key_as_string(key):
    """For keys, we don't want the key to appear in INFO log outputs.
    Instead, we display the first 4 chars and the last 4 chars.
    """
    return (key[:4] + "*******" + key[-4:]) if isinstance(key, string_types) else str(key)


def get_cplex_version(context):
    '''Submits a minimal job on DOCplexcloud and get the CPLEX version from
    logs
    Args:
        context: The context to use to connect
    Returns:
        A tuple (major, minor, micro), for instance (12, 7, 1)
    '''
    import re
    from docplex.mp.compat23 import StringIO
    from docplex.mp.model import Model
    logs = StringIO()
    model = Model()
    model.solve(url=context.solver.docloud.url,
                key=context.solver.docloud.key,
                log_output=logs)
    v = re.search('CPLEX version ([0-9][0-9])([0-9][0-9])([0-9][0-9])([0-9][0-9])', logs.getvalue())
    try:
        major = int(v.group(1))
        minor = int(v.group(2))
        micro = int(v.group(3))
        return (major, minor, micro)
    except IndexError:
        raise RuntimeError('Could not find CPLEX version in resulting logs')
    except:
        raise ValueError('Could not determine version from %s' % v.group(0))


class DOcloudConnectorException(Exception):
    def __init__(self, msg):
        Exception.__init__(self, msg)
        self.message = msg


class DOcloudEvent(object):
    """Used internally to manage events that must be processed in
    the main thread.

    The Progress and log monitor loop poll the service for progress and log
    information, then queue DOcloudEvents on an internal queue. That queue
    is processed by a loop in the main thread, making sure that logs are
    printed in the main thread and progress listeners are called in the
    main thread too.
    """
    def __init__(self, event_type, data):
        """The type if a string. Can be "log" or "progress"
        """
        self.type = event_type
        self.data = data


class DOcloudConnector(object):

    def __init__(self, docloud_context, log_output=None):
        """ Starts a connector which URL and authorization are stored in the
        specified context, along with other connection parameters

        The `docloud_context` refers to the context.solver.docloud node of a
        context.

        Args:
            log_output: The log output stream
        """
        if docloud_context is None or not has_credentials(docloud_context):
            raise DOcloudConnectorException("Please provide DOcplexcloud credentials")

        # store this for future usage
        self.docloud_context = docloud_context

        url = docloud_context.url
        auth = docloud_context.key

        self.logger = get_logger('DOcloudConnector', self.docloud_context.verbose)

        self.logger.info("DOcplexcloud connection using url = " + str(url) + " api_key = " + key_as_string(auth))
        self.logger.info("DOcplexcloud SSL verification = " + str(docloud_context.verify))

        self.logger.info("   waittime = " + str(docloud_context.waittime))
        self.logger.info("   timeout = " + str(docloud_context.timeout))

        self.jobInfo = None
        self.run_deterministic = docloud_context.run_deterministic
        self.log_output = log_output

        self.timed_out = False
        self.results = {}

    def log(self, msg, *args):
        if self.docloud_context.verbose:
            log_msg = "* {0}".format(resolve_pattern(msg, args))
            self.logger.info(log_msg)
            if self.log_output is not None:
                self.log_output.write(log_msg)

    def _as_string(self, content):
        resp_content_as_string = content
        if not isinstance(resp_content_as_string, str):
            resp_content_as_string = content.decode('utf-8')
        return resp_content_as_string

    def submit_model_data(self,
                          attachments=None,
                          gzip=False,
                          info_callback=None,
                          info_to_monitor=None):
        """Submits a job to the cloud service.

        Args:
            attachments: A list of attachments. Each attachement is a dict with
                the following keys:
                   - 'name' : the name of the attachment
                   - 'data' : the data for the attachment
            gzip: If ``True``, data is gzipped before sent over the network
            info_callback: A call back to be called when some info are available.
                That callback takes one parameter that is a dict containing
                the info as they are available.
            info_to_monitor: A set of information to monitor with info_callback.
                Currently, can be ``jobid`` and ``progress``.
        """
        self.__vars = None
        self.timed_out = False
        self.results.clear()

        if not info_to_monitor:
            info_to_monitor = {}

        # check that url is valid
        parts = urlparse(self.docloud_context.url)
        if not parts.scheme:
            raise DOcloudConnectorException("Malformed URL: '%s': No schema supplied." % self.docloud_context.url)

        proxies = self.docloud_context.proxies
        try:
            client = JobClient(self.docloud_context.url,
                               self.docloud_context.key,
                               proxies=proxies)
        except TypeError:
            # docloud client <= 1.0.172 do not have the proxes
            warnings.warn("Using a docloud client that do not support warnings in init()",
                          UserWarning)
            client = JobClient(self.docloud_context.url,
                               self.docloud_context.key)
        self.log("client created")
        if proxies:
            self.log("proxies = %s" % proxies)

        # prepare client
        if self.docloud_context.log_requests:
            client.rest_callback = \
                lambda m, u, *a, **kw: self._rest_callback(m, u, *a, **kw)
        client.verify = self.docloud_context.verify
        client.timeout = self.docloud_context.get('timeout', None)

        try:
            try:
                # Extract the list of attachment names
                att_names = [a['name'] for a in attachments]

                # create job
                jobid = client.create_job(attachments=att_names,
                                          parameters=self.docloud_context.job_parameters)
                self.log("job creation submitted, id is: {0!s}".format(jobid))
                if info_callback and 'jobid' in info_to_monitor:
                    info_callback({'jobid': jobid})
            except ConnectionError as c_e:
                raise DOcloudConnectorException("Cannot connect to {0}, error: {1}".format(self.docloud_context.url, str(c_e)))

            try:
                # now upload data
                for a in attachments:
                    pos = 0
                    if 'data' in a:
                        att_data = {'data': a['data']}
                    elif 'file' in a:
                        att_data = {'file': a['file']}
                        pos = a['file'].tell()
                    elif 'filename' in a:
                        att_data = {'filename': a['filename']}

                    client.upload_job_attachment(jobid,
                                                 attid=a['name'],
                                                 **att_data)
                    self.log("Attachment: %s has been uploaded" % a['name'])
                    if self.docloud_context.debug_dump_dir:
                        target_dir = self.docloud_context.debug_dump_dir
                        if not os.path.exists(target_dir):
                            os.makedirs(target_dir)
                        self.log("Dumping input attachment %s to dir %s" % (a['name'], target_dir))
                        with open(os.path.join(target_dir, a['name']), "wb") as f:
                            if 'data' in 'a':
                                if isinstance(a['data'], bytes):
                                    f.write(a['data'])
                                else:
                                    f.write(a['data'].encode('utf-8'))
                            else:
                                a['file'].seek(pos)
                                f.write(a['file'])
                # execute job
                client.execute_job(jobid)
                self.log("DOcplexcloud execute submitted has been started")
                # get job execution status until it's processed or failed
                timedout = False
                try:
                    self._executionStatus = self.wait_for_completion(client,
                                                                     jobid,
                                                                     info_callback=info_callback,
                                                                     info_to_monitor=info_to_monitor)
                except DOcloudInterruptedException:
                    timedout = True
                self.log("docloud execution has finished")
                # get job status. Do this before any time out handling
                self.jobInfo = client.get_job(jobid)

                if self.docloud_context.fire_last_progress and info_callback:
                    progress_data = self.map_job_info_to_progress_data(self.jobInfo)
                    info_callback({'progress': progress_data})

                if timedout:
                    self.timed_out = True
                    self.log("Solve timed out after {waittime} sec".format(waittime=self.docloud_context.waittime))
                    return
                # get solution => download all attachments
                try:
                    for a in client.get_job_attachments(jobid):
                        if a['type'] == 'OUTPUT_ATTACHMENT':
                            name = a['name']
                            self.log("Downloading attachment '%s'" % name)
                            attachment_as_string = self._as_string(client.download_job_attachment(jobid,
                                                                                                  attid=name))
                            self.results[name] = attachment_as_string
                            if self.docloud_context.debug_dump_dir:
                                target_dir = self.docloud_context.debug_dump_dir
                                if not os.path.exists(target_dir):
                                    os.makedirs(target_dir)
                                self.log("Dumping attachment %s to dir %s" % (name, target_dir))
                                with open(os.path.join(target_dir, name), "wb") as f:
                                    f.write(attachment_as_string.encode('utf-8'))
                except DOcloudNotFoundError:
                    self.log("no solution in attachment")
                self.log("docloud results have been received")
                # on_solve_finished_cb
                if self.docloud_context.on_solve_finished_cb:
                    self.docloud_context.on_solve_finished_cb(jobid=jobid,
                                                              client=client,
                                                              connector=self)
                return
            finally:
                if self.docloud_context.delete_job:
                    deleted = client.delete_job(jobid)
                    self.log("delete status for job: {0!s} = {1!s}".format(jobid, deleted))

        finally:
            client.close()

    def wait_for_completion(self, client, jobid,
                            info_callback=None, info_to_monitor=None):
        def status_poll(loop):
            """Callback to check execution status and stop the loop as soon
            as the job is finished. The loop is stopped as soon as the status
            is for a finished job,
            """
            status = loop.client.get_execution_status(loop.jobid)
            if JobExecutionStatus.isEnded(status):
                loop.status = status
                loop.stop()

        def waittime_timeout(loop):
            """Callback to stop completly the loop once the max waittime
            has been hit.
            """
            loop.stop()
            loop.timed_out = True

        def download_logs(loop, using_threads, log_output):
            """Function/Callback to download logs.

            We will use that function in threads (log_output will be None),
            were we want log events to be put in the loop's event_queue,
            and also after the loop has stopped(), to dump any log items
            remaining on the server (in this case, log_output will
            be the stream to write logs to)

            Args:
                loop: the loop
                using_threads: if true, we are using threads and should queue
                    events so that they are processed in the main thread
                log_output: the log_output
            """
            logs = loop.client.get_log_items(loop.jobid, loop.last_seqid, True)
            for log in logs:
                loop.last_seqid = log['seqid'] + 1
                for r in log['records']:
                    level = r['level'][:4]
                    date = r['date']
                    message = r['message'].rstrip()
                    d = datetime.utcfromtimestamp(int(float(date) * 0.001))
                    m = "[{date}Z, {level}] {message}\n".format(date=d.isoformat(),
                                                                level=level,
                                                                message=message
                                                                )
                    if using_threads:
                        loop.event_queue.put(DOcloudEvent("log", m))
                    else:
                        log_output.write(m)

        def progress_poll(loop, using_threads, info_callback, info_to_monitor):
            """Function/Callback to poll and download progress.

            That function polls the service for progress, then queue
            progress events in the main loop.

            Args:
                loop: the loop
                using_threads: if true, we are using threads and should queue
                    events so that they are processed in the main thread
                info_callback: the info callback
                info_to_monitor: what info does the callback want to monitor
            """
            if 'progress' in info_to_monitor and info_callback:
                logger = self.docloud_context.verbose_progress_logger
                if logger:
                    logger.info("polling progress")
                info = loop.client.get_job(loop.jobid)
                if logger:
                    logger.info("job info: %s" % json.dumps(info, indent=3))
                if 'details' in info:  # there are some info available
                    if logger:
                        logger.info("generating progress event using_threads = %s" % using_threads)
                    progress_data = self.map_job_info_to_progress_data(info)
                    if using_threads:
                        loop.event_queue.put(DOcloudEvent("progress", progress_data))
                    else:
                        info_callback({'progress': progress_data})

        class JobMonitor(CyclicLoop):
            """A cyclic loop with some encapsuled data"""
            def __init__(self, client, jobid, log_output):
                super(JobMonitor, self).__init__()
                self.client = client
                self.jobid = jobid
                self.log_output = log_output

                self.status = None
                self.timed_out = False
                self.last_seqid = 0

        if not info_to_monitor:
            info_to_monitor = {}

        # interval to check job status
        status_poll_interval = client.nice
        log_poll_interval = self.docloud_context.get('log_poll_interval')
        if log_poll_interval is None:
            log_poll_interval = client.nice * 3
        progress_poll_interval = self.docloud_context.get('progress_poll_interval')
        if progress_poll_interval is None:
            progress_poll_interval = client.nice * 3

        # The cyclic loop
        loop = JobMonitor(client, jobid, self.log_output)
        using_threads = False

        # configure status log event
        loop.enter(status_poll_interval, 1, status_poll, (loop, ))
        # configure log poll event
        if self.log_output:
            self.log("Polling logs every %s sec" % log_poll_interval)
            loop.enter(log_poll_interval, 1, download_logs,
                       (loop, using_threads, self.log_output))
        # configure progress poll event
        if 'progress' in info_to_monitor:
            self.log("Polling progress every %s sec" % progress_poll_interval)
            loop.enter(progress_poll_interval, 1,
                       progress_poll,
                       (loop, using_threads, info_callback, info_to_monitor))

        # If there's a waittime, configure an event to stop the loop after
        # ``waittime``
        if self.docloud_context.waittime:
            loop.enter(self.docloud_context.waittime, 1, waittime_timeout, (loop, ))
            self.log("waiting for job completion with a wait time of {waittime} sec".format(waittime=self.docloud_context.waittime))
        else:
            self.log("waiting for job completion with no wait time")

        # we want the dumps of the log to happen in the main thread
        # we also want the progress listener events to come in the main thread
        # this function is guaranteed to run in the main thread by the loop
        def main_thread_worker(event):
            if event.type == "log":
                loop.log_output.write(event.data)
            elif event.type == "progress":
                if info_callback:
                    info_callback({'progress': event.data})

        kwargs = {}
        if using_threads:
            kwargs['mt_worker'] = main_thread_worker
        loop.start(**kwargs)

        if self.log_output:
            # this will download the log items that were not downloaded in
            # the loop. Using using_threads=False to force a dump of the log
            # output
            download_logs(loop, False, self.log_output)

        if loop.timed_out:
            self.log("Job Timed out")
            raise DOcloudInterruptedException("Timeout after {0}".format(self.docloud_context.waittime), jobid=jobid)

        return loop.status

    def get_cplex_details(self):
        if self.jobInfo:
            return self.jobInfo.get("details")

    def get_solve_status(self):
        if 'solveStatus' in self.jobInfo:
            return JobSolveStatus[self.jobInfo['solveStatus']]
        else:
            return None

    def _rest_callback(self, method, url, *args, **kwargs):
        """The callback called by the DOcplexcloud client to log REST operations
        """
        self.logger.info("{0} {1}".format(method, url))
        if len(args) > 0:
            self.logger.info("   Additionnal args : {0}".format(','.join(args)))
        for k, v in iteritems(kwargs):
            self.logger.info("   {0}: {1}".format(k, v))

    def map_job_info_to_progress_data(self, info):
        """ Map job info as downloaded from the cplex cloud worker to
        docplex.mp.progress.ProgressData

        Args:
            info: The info as a dict
        Returns:
            A ProgressData
        """
        from docplex.mp.progress import ProgressData
        has_incumbent = current_objective = best_bound = mip_gap = current_nb_iterations = current_nb_nodes = remaining_nb_nodes = 0
        time = 0
        details = info.get('details')
        if details:
            current_objective = float(details.get('PROGRESS_CURRENT_OBJECTIVE', 0))
            best_bound = float(details.get('PROGRESS_BEST_OBJECTIVE', 0))
            mip_gap = None
            if 'PROGRESS_CURRENT_OBJECTIVE' in details and 'PROGRESS_BEST_OBJECTIVE' in details:
                if current_objective > 0:
                    mip_gap = abs(current_objective - best_bound) / current_objective
            current_nb_nodes = int(details.get('cplex.nodes.processed', 0))
            remaining_nb_nodes = int(details.get('cplex.nodes.left', 0))
            # assume that there's an incubent if ther's a gap
            has_incumbent = 'PROGRESS_GAP' in details
        if 'startedAt' in info and 'updatedAt' in info:
            time = ((info.get('updatedAt')) - int(info.get('startedAt'))) / 1000
        pg = ProgressData(0, has_incumbent, current_objective, best_bound, mip_gap,
                          0, current_nb_nodes, remaining_nb_nodes,
                          time, 0)
        return pg



# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2015, 2016
# --------------------------------------------------------------------------

from io import BytesIO

import os

from six import iteritems
import tempfile

import warnings

from docplex.mp.engine import IndexerEngine

try:
    # import DOcloudConnector only if JobClient is available
    from docloud.job import JobClient
    from docplex.mp.docloud_connector import DOcloudConnector
except ImportError as ie:
    # just ignore if docloud is not available
    pass

from docplex.mp.internal.json_solution_handler import JSONSolutionHandler
from docplex.mp.internal.json_infeasibility_handler import JSONInfeasibilityHandler
from docplex.mp.internal.json_conflict_handler import JSONConflictHandler
from docplex.mp.printer_factory import ModelPrinterFactory
from docplex.mp.solution import SolveSolution, SolutionMSTPrinter
from docplex.mp.anno import ModelAnnotationPrinter
from docplex.mp.sdetails import SolveDetails
from docplex.mp.utils import DOcplexException, make_path
from docplex.mp.utils import normalize_basename
from docplex.mp.constants import ConflictStatus
from docplex.mp.conflict_refiner import TConflictConstraint, VarLbConstraintWrapper, VarUbConstraintWrapper, ConflictRefinerResult
from docplex.mp.constr import LinearConstraint, QuadraticConstraint, IndicatorConstraint

from docplex.util.environment import make_attachment_name

from docplex.mp.format import LP_format
from docplex.mp.compat23 import StringIO

from xml.etree.ElementTree import Element, SubElement
from xml.etree import ElementTree
from xml.dom import minidom

from collections import defaultdict
import numbers
import sys

# gendoc: ignore

# this is the default exchange format for docloud
_DEFAULT_EXCHANGE_FORMAT = LP_format

# default attachment name to trigger conflict refiner
feasibility_name = "file.feasibility"


class FeasibilityPrinter(object):
    extension = ".feasibility"

    @classmethod
    def print_to_stream(cls, relaxables, out, extension=extension):
        if out is None:
            # prints on standard output
            cls.print_internal(sys.stdout, relaxables)

        elif isinstance(out, str):
            # a string is interpreted as a path name
            path = out if out.endswith(extension) else out + extension
            with open(path, "w") as of:
                cls.print_to_stream(relaxables, of)
        else:
            try:
                cls.print_internal(out, relaxables)

            except AttributeError:  # pragma: no cover
                pass  # pragma: no cover
                # stringio will raise an attribute error here, due to with
                # print("Cannot use this an output: %s" % str(out))

    @classmethod
    def print_internal(cls, out, relaxables):
        # INTERNAL: out is resolved here to a writeable stream...
        # the infeasibilityFile flag is here to trigger the output of infeasibilities
        # -> not the default.
        out.write("<CPLEXFeasopt infeasibilityFile='true'>\n")
        out.write("  <rhs>\n")
        for relaxable_group in relaxables:
            pref, cts = relaxable_group
            for ct in cts:
                out.write("    <relax index=\"{0}\" preference=\"{1}\"/>".format(ct.index, pref))
                ctname = ct.name
                if ctname:
                    out.write("  <!-- {0} -->".format(ctname))
                out.write("\n")

        out.write("  </rhs>\n")
        out.write("</CPLEXFeasopt>\n")


# constraint_type_by_ct_type = {'lb': VarLbConstraintWrapper,
#                               'ub': VarUbConstraintWrapper,
#                               'lin': LinearConstraint,
#                               'quad': QuadraticConstraint,
#                               'ind': IndicatorConstraint,
#                               'sos': None}


ct_type_by_constraint_type = {VarLbConstraintWrapper: 'lb',
                              VarUbConstraintWrapper: 'ub',
                              LinearConstraint: 'lin',
                              QuadraticConstraint: 'quad',
                              IndicatorConstraint: 'ind'}


class ConflictRefinerPrinter(object):
    extension = ".feasibility"

    @classmethod
    def print_to_stream(cls, artifact_as_xml, out, extension=extension):
        if out is None:
            # prints on standard output
            cls.print_internal(sys.stdout, artifact_as_xml)

        elif isinstance(out, str):
            # a string is interpreted as a path name
            path = out if out.endswith(extension) else out + extension
            with open(path, "w") as of:
                cls.print_to_stream(artifact_as_xml, of)
        else:
            try:
                cls.print_internal(out, artifact_as_xml)

            except AttributeError:  # pragma: no cover
                pass  # pragma: no cover
                # stringio will raise an attribute error here, due to with
                # print("Cannot use this an output: %s" % str(out))

    @classmethod
    def print_internal(cls, out, artifact_as_xml):
        ElementTree.ElementTree(artifact_as_xml.get_root()).write(out, 'utf-8')


class CPLEXRefineConflictExtArtifact(object):
    '''
    XML builder for DOcplexcloud artifact to trigger Conflict Refiner execution
    '''

    def __init__(self):
        self._root = Element('CPLEXRefineconflictext')
        self._grps_index = {}
        self._grps_dict = defaultdict(list)

    def get_root(self):
        return self._root

    def add_constraints(self, ct_type, elements, preference=1.0, preference_dict=None):
        for (elem, name) in elements:
            pref = preference
            if preference_dict is not None:
                new_pref = preference_dict.get(elem, None)
                if new_pref is not None and isinstance(new_pref, numbers.Number):
                    pref = new_pref
            group = self.add_group(pref)
            con = SubElement(group, "con")
            con.set("name", name)
            con.set("type", ct_type)
            self._grps_dict[self._grps_index[group]].append(elem)

    def add_group(self, preference):
        group = SubElement(self._root, "group")
        group.set("preference", repr(preference))
        self._grps_index[group] = len(self._grps_index)
        return group

    def add_constraint_to_group(self, group, ct):
        con = SubElement(group, "con")
        con.set("name", ct.name)
        constraint_type = type(ct)
        con.set("type", ct_type_by_constraint_type.get(constraint_type, ""))
        self._grps_dict[self._grps_index[group]].append(ct)

    def tostring(self, encoding='utf-8'):
        """Return a string for the XML Element hierarchy.
        """
        return ElementTree.tostring(self._root, encoding)

    def pretty_print(self):
        """Return a pretty-printed XML string for the Element.
        """
        as_string = ElementTree.tostring(self._root, 'utf-8')
        dom = minidom.parseString(as_string)
        return dom.toprettyxml(indent="  ")


# noinspection PyProtectedMember
class DOcloudEngine(IndexerEngine):
    """ Engine facade stub to defer solve to drop-solve URL
    """

    def _print_feasibility(self, out, relaxables):
        pass

    def get_cplex(self):
        raise DOcplexException("{0} engine contains no instance of CPLEX".format(self.name))

    def __init__(self, mdl, exchange_format=None, **kwargs):
        IndexerEngine.__init__(self)

        warnings.warn(
            'Solve using \'docloud\' agent is deprecated. Consider submitting your model to DOcplexcloud. See https://ibm.biz/BdYhhK',
            DeprecationWarning)

        docloud_context = kwargs.get('docloud_context')
        # --- log output can be overridden at solve time, so use te one from the context, not the model's
        actual_log_output = kwargs.get('log_output') or mdl.log_output

        self._model = mdl
        self._connector = DOcloudConnector(docloud_context, log_output=actual_log_output)
        self._exchange_format = exchange_format or docloud_context.exchange_format or _DEFAULT_EXCHANGE_FORMAT

        mangle_names = mdl.ignore_names or mdl.context.solver.docloud.mangle_names
        self._printer = ModelPrinterFactory.new_printer(self._exchange_format, full_obj=True)
        if mangle_names:
            self._printer.set_mangle_names(True)

        # -- results.
        self._lpname_to_var_map = {}
        self._solve_details = SolveDetails.make_dummy()
        self._quality_metrics = {}  # latest quality metrics from json

        # noinspection PyPep8
        self.debug_dump = docloud_context.debug_dump
        self.debug_dump_dir = docloud_context.debug_dump_dir

    def _new_printer(self, ctx):
        return self._printer

    def supports_logical_constraints(self):
        # <-> is supposed to be supported in LP?
        return True, None

    docloud_solver_name = 'cplex_cloud'

    @property
    def name(self):
        return self.docloud_solver_name

    def can_solve(self):
        """
        :return: true, as this solver can solve!
        """
        return True

    def connect_progress_listeners(self, progress_listener_list, model):
        if progress_listener_list:
            self._model.warning("Progress listeners are not supported on DOcplexcloud.")

    def register_callback(self, cb):
        self._model.fatal('Callbacks are not available on DOcplexcloud')

    @staticmethod
    def _docloud_cplex_version():
        # INTERNAL: returns the version of CPLEX used in DOcplexcloud
        # for now returns a string. maybe we could ping Docloud and get a dynamic answer.
        return "12.6.3.0"

    def _serialize_parameters(self, parameters, write_level=None, relax_mode=None):
        # return a string in PRM format
        # overloaded params are:
        # - THREADS = 1, if deterministic, else not mentioned.
        # the resulting string will contain all non-default parameters,
        # AND those overloaded.
        # No side effect on actual model parameters
        overloaded_params = dict()
        if relax_mode is not None:
            overloaded_params[parameters.feasopt.mode] = relax_mode.value

        # Do not override write level anymore
        if write_level is not None:
            overloaded_params[parameters.output.writelevel] = write_level

        if self._connector.run_deterministic:
            # overloaded_params[mdl_parameters.threads] = 1 cf RTC28458
            overloaded_params[parameters.parallel] = 1  # 1 is deterministic

        # do we need to limit the version to the one use din docloud
        # i.e. if someone has a *newer* version than docloud??
        prm_data = parameters.export_prm_to_string(overloaded_params)
        return prm_data

    def serialize_model_as_file(self, mdl):
        # step 1 : prints the model in whatever exchange format
        printer = self._new_printer(ctx=mdl.context)

        if self._exchange_format.is_binary:
            filemode = "w+b"
        else:
            filemode = "w+"

        lp_output = tempfile.NamedTemporaryFile(mode=filemode, delete=False)

        printer.printModel(mdl, lp_output)

        # lp name to docplex var
        self._lpname_to_var_map = printer.get_name_to_var_map(mdl)

        # DEBUG: dump request file
        if self.debug_dump:
            dump_path = make_path(error_handler=mdl.error_handler,
                                  basename=mdl.name,
                                  output_dir=self.debug_dump_dir,
                                  extension=printer.extension(),
                                  name_transformer="docloud_%s")
            print("DEBUG DUMP in " + dump_path)
            with open(dump_path, filemode) as out_file:
                lp_output.seek(0)
                out_file.write(lp_output)

        lp_output.close()
        return lp_output.name

    def serialize_model(self, mdl):
        # step 1 : prints the model in whatever exchange format
        printer = self._new_printer(ctx=mdl.context)

        if self._exchange_format.is_binary:
            filemode = "wb"
            oss = BytesIO()
        else:
            filemode = "w"
            oss = StringIO()

        printer.printModel(mdl, oss)

        # lp name to docplex var
        self._lpname_to_var_map = printer.get_name_to_var_map(mdl)

        # DEBUG: dump request file
        if self.debug_dump:
            dump_path = make_path(error_handler=mdl.error_handler,
                                  basename=mdl.name,
                                  output_dir=self.debug_dump_dir,
                                  extension=printer.extension(),
                                  name_transformer="docloud_%s")
            print("DEBUG DUMP in " + dump_path)
            with open(dump_path, filemode) as out_file:
                out_file.write(oss.getvalue())

        if self._exchange_format.is_binary:
            model_data = oss.getvalue()
        else:
            model_data = oss.getvalue().encode('utf-8')
        return model_data

    def _serialize_relaxables(self, relaxables):
        oss = StringIO()
        FeasibilityPrinter.print_to_stream(out=oss, relaxables=relaxables)
        serialized_relaxables = oss.getvalue()
        return serialized_relaxables

    def _dump_if_required(self, data, mdl, basename, extension, is_binary=False, forced=False):
        # INTERNAL
        if self.debug_dump or forced:
            relax_path = make_path(error_handler=mdl.error_handler,
                                   basename=basename,
                                   output_dir=self.debug_dump_dir,
                                   extension=extension,
                                   name_transformer=None)
            if isinstance(data, bytes) or isinstance(data, bytearray):
                is_binary = True  # For binary when data is binary
            fmode = "wb" if is_binary else "w"
            with open(relax_path, fmode) as out_file:
                out_file.write(data)

    def _make_attachment(self, attachment_name, attachment_data):
        # INTERNAL
        return {'name': attachment_name, 'data': attachment_data}

    def solve_relaxed(self, mdl, prio_name, relaxable_groups, relax_mode, parameters=None):
        # --- 1 serialize
        job_name = normalize_basename(mdl.name, force_lowercase=True)
        model_data = self.serialize_model(mdl)
        docloud_parameters = mdl.parameters
        prm_data = self._serialize_parameters(docloud_parameters, write_level=1, relax_mode=relax_mode)
        prm_name = self._make_attachment_name(job_name, '.prm')
        feasopt_data = self._serialize_relaxables(relaxable_groups)

        # --- dump if need be
        if prio_name:
            prio_name = "_%s" % prio_name
        relax_basename = normalize_basename("%s_feasopt%s" % (mdl.name, prio_name))

        prm_basename = normalize_basename("%s_feasopt" % mdl.name)
        self._dump_if_required(model_data, mdl, basename=job_name, extension=".lp", forced=True)
        self._dump_if_required(feasopt_data, mdl, basename=relax_basename, extension=FeasibilityPrinter.extension,
                               forced=True)
        self._dump_if_required(prm_data, mdl, basename=prm_basename, extension=".prm", forced=True)

        # --- submit job somehow...
        attachments = []
        model_name = normalize_basename(job_name) + self._exchange_format.extension
        attachments.append(self._make_attachment(model_name, model_data))

        attachments.append(self._make_attachment(prm_name, prm_data))
        attachments.append(self._make_attachment(normalize_basename(job_name) + FeasibilityPrinter.extension,
                                                 feasopt_data))

        # here we go...
        def notify_info(info):
            if "jobid" in info:
                mdl.fire_jobid(jobid=info["jobid"])
            if "progress" in info:
                mdl.fire_progress(progress_data=info["progress"])

        connector = self._connector
        mdl.notify_start_solve()
        connector.submit_model_data(attachments,
                                    gzip=not self._exchange_format.is_binary,
                                    info_callback=notify_info,
                                    info_to_monitor={'jobid', 'progress'})

        # --- cplex solve details
        json_details = connector.get_cplex_details()
        self._solve_details = SolveDetails.from_json(json_details)
        # ---

        # --- build a solution object, or None
        solution_handler = JSONSolutionHandler(connector.results.get('solution.json'))
        if not solution_handler.has_solution:
            mdl.notify_solve_failed()
            return None
        else:
            infeas_json = connector.results.get('infeasibilities.json')
            infeas_handler = JSONInfeasibilityHandler(infeas_json) if infeas_json else None
            sol = self._make_relaxed_solution(mdl, solution_handler, infeas_handler)
            return sol

    def _serialize_conflict_refiner(self, artifact_as_xml):
        if sys.version_info[0] < 3:
            oss = StringIO()
        else:
            oss = BytesIO()
        ConflictRefinerPrinter.print_to_stream(artifact_as_xml, out=oss)
        serialized_artifact = oss.getvalue()
        return serialized_artifact

    def refine_conflict(self, mdl, preferences=None, groups=None, parameters=None):
        """ Starts conflict refiner on the model.

        Args:
            mdl: The model for which conflict refinement is performed.
            preferences: a dictionary defining constraints preferences.
            groups: a list of ConstraintsGroup.
            :parameters: cplex parameters .

        Returns:
            A list of "TConflictConstraint" namedtuples, each tuple corresponding to a constraint that is
            involved in the conflict.
            The fields of the "TConflictConstraint" namedtuple are:
                - the name of the constraint or None if the constraint corresponds to a variable lower or upper bound
                - a reference to the constraint or to a wrapper representing a Var upper or lower bound
                - an :enum:'docplex.mp.constants.ConflictStatus' object that indicates the
                conflict status type (Excluded, Possible_member, Member...)
            This list is empty if no conflict is found by the conflict refiner.
        """
        # Before submitting the job, we will build the list of attachments
        attachments = []

        # make sure model is the first attachment: that will be the name of the job on the console
        job_name = "python_%s" % self._model.name
        model_data = self.serialize_model(self._model)
        mprinter = self._new_printer(ctx=mdl.context)
        model_name = normalize_basename(job_name) + self._exchange_format.extension
        attachments.append({'name': model_name, 'data': model_data})

        # Conflict Refiner Ext
        artifact_as_xml = CPLEXRefineConflictExtArtifact()
        if groups is None or groups == []:
            # Add all constraints
            artifact_as_xml.add_constraints(ct_type_by_constraint_type[VarUbConstraintWrapper],
                                            [(VarUbConstraintWrapper(v), mprinter._var_print_name(v))
                                             for v in self._model.iter_variables()],
                                            preference=1.0)
            artifact_as_xml.add_constraints(ct_type_by_constraint_type[VarLbConstraintWrapper],
                                            [(VarLbConstraintWrapper(v), mprinter._var_print_name(v))
                                             for v in self._model.iter_variables()],
                                            preference=1.0)
            artifact_as_xml.add_constraints(ct_type_by_constraint_type[LinearConstraint],
                                            [(c, mprinter.linearct_print_name(c))
                                             for c in self._model.iter_linear_constraints()],
                                            preference_dict=preferences)
            artifact_as_xml.add_constraints(ct_type_by_constraint_type[QuadraticConstraint],
                                            [(c, mprinter.qc_print_name(c))
                                             for c in self._model.iter_quadratic_constraints()],
                                            preference_dict=preferences)
            artifact_as_xml.add_constraints(ct_type_by_constraint_type[IndicatorConstraint],
                                            [(c, mprinter.logicalct_print_name(c))
                                             for c in self._model.iter_indicator_constraints()],
                                            preference_dict=preferences)
        else:
            for grp in groups:
                group = artifact_as_xml.add_group(grp.preference)
                for ct in grp._cts:
                    artifact_as_xml.add_constraint_to_group(group, ct)

        conflict_refiner_data = self._serialize_conflict_refiner(artifact_as_xml)

        attachments.append({'name': feasibility_name, 'data': conflict_refiner_data})

        def notify_info(info):
            if "jobid" in info:
                self._model.fire_jobid(jobid=info["jobid"])
            if "progress" in info:
                self._model.fire_progress(progress_data=info["progress"])

        # This block used to be try/catched for DOcloudConnector exceptions
        # and DOcloudException, but then infrastructure error were not
        # handled properly. Now we let the exception raise.
        connector = self._connector
        self._model.notify_start_solve()
        connector.submit_model_data(attachments,
                                    gzip=not self._exchange_format.is_binary,
                                    info_callback=notify_info,
                                    info_to_monitor={'jobid', 'progress'})

        # --- build a conflict object, or None
        conflicts_handler = JSONConflictHandler(connector.results.get('conflict.json'), artifact_as_xml._grps_dict)
        if not conflicts_handler.has_conflict:
            return []
        else:
            return self._get_conflicts_cloud(conflicts_handler)
        # ---

    def _get_conflicts_cloud(self, conflicts_handler):
        conflict_grps = conflicts_handler.get_conflict_grps_list()
        conflicts = []

        for index, elems, c_status in conflict_grps:
            if c_status is None:
                self._model.error("Undefined status for constraint conflict for group index: {0}".format(index))
                continue
            if c_status == ConflictStatus.Excluded:
                continue

            for elem in elems:
                if isinstance(elem, VarLbConstraintWrapper):
                    conflicts.append(TConflictConstraint(None, elem, c_status))

                elif isinstance(elem, VarUbConstraintWrapper):
                    conflicts.append(TConflictConstraint(None, elem, c_status))

                else:
                    conflicts.append(TConflictConstraint(elem.name, elem, c_status))

        return ConflictRefinerResult(conflicts, refined_by=self.name)

    def _make_attachment_name(self, basename, extension):
        return make_attachment_name(basename + extension)

    def export_one_mip_start(self, mipstart, job_name, attachments):
        warmstart_data = SolutionMSTPrinter.print_to_string(mipstart).encode('utf-8')
        mipstart_name = mipstart.name.lower() if mipstart.name else job_name
        warmstart_name = self._make_attachment_name(mipstart_name, ".mst")
        attachments.append({'name': warmstart_name, 'data': warmstart_data})

    # noinspection PyProtectedMember
    def solve(self, mdl, parameters=None, **kwargs):
        # Before submitting the job, we will build the list of attachments
        # parameters are CPLEX parameters
        lex_mipstart = kwargs.pop('_lex_mipstart', None)
        attachments = []

        # make sure model is the first attachment: that will be the name of the job on the console
        job_name = normalize_basename("python_%s" % mdl.name)
        model_file = self.serialize_model_as_file(mdl)
        try:
            model_data_name = self._make_attachment_name(job_name, self._exchange_format.extension)
            attachments.append({'name': model_data_name, 'filename': model_file})

            # prm
            docloud_parameters = parameters if parameters is not None else mdl.parameters
            prm_data = self._serialize_parameters(docloud_parameters)
            prm_name = self._make_attachment_name(job_name, '.prm')
            attachments.append({'name': prm_name, 'data': prm_data})

            # warmstart_data
            # export mipstart solution in CPLEX mst format, if any, else None
            # if within a lexicographic solve, th elex_mipstart supersedes allother mipstarts
            if lex_mipstart:
                mipstart_name = lex_mipstart.name.lower() if lex_mipstart.name else job_name
                warmstart_data = SolutionMSTPrinter.print_to_string(lex_mipstart).encode('utf-8')
                warmstart_name = self._make_attachment_name(mipstart_name, ".mst")
                attachments.append({'name': warmstart_name, 'data': warmstart_data})

            elif mdl.number_of_mip_starts:
                mipstart_name = job_name
                warmstart_name = self._make_attachment_name(mipstart_name, ".mst")
                mdl_mipstarts = [s for s, _ in mdl.iter_mip_starts()]
                mdl_efforts = [eff for (_, eff) in mdl.iter_mip_starts()]
                warmstart_data = SolutionMSTPrinter.print_to_string(mdl_mipstarts, effort_level=mdl_efforts,
                                                                    use_lp_names=True).encode('utf-8')
                attachments.append({'name': warmstart_name, 'data': warmstart_data})

            # benders annotation
            if mdl.has_benders_annotations():
                anno_data = ModelAnnotationPrinter.print_to_string(mdl).encode('utf-8')
                anno_name = self._make_attachment_name(job_name, '.ann')
                attachments.append({'name': anno_name, 'data': anno_data})

            # info_to_monitor = {'jobid'}
            # if mdl.progress_listeners:
            # info_to_monitor.add('progress')

            def notify_info(info):
                if "jobid" in info:
                    mdl.fire_jobid(jobid=info["jobid"])
                if "progress" in info:
                    mdl.fire_progress(progress_data=info["progress"])

            # This block used to be try/catched for DOcloudConnector exceptions
            # and DOcloudException, but then infrastructure error were not
            # handled properly. Now we let the exception raise.
            connector = self._connector
            mdl.notify_start_solve()
            connector.submit_model_data(attachments,
                                        gzip=not self._exchange_format.is_binary,
                                        info_callback=notify_info,
                                        info_to_monitor={'jobid', 'progress'})

            # --- cplex solve details
            json_details = connector.get_cplex_details()
            self._solve_details = SolveDetails.from_json(json_details)
            self._solve_details._quality_metrics = self._compute_quality_metrics(json_details)
            # ---

            # --- build a solution object, or None
            solution_handler = JSONSolutionHandler(connector.results.get('solution.json'))
            if not solution_handler.has_solution:
                mdl.notify_solve_failed()
                solution = None
            else:
                solution = self._make_solution(mdl, solution_handler)
            # ---

            return solution
        finally:
            if os.path.isfile(model_file):
                os.remove(model_file)

    def _var_by_cloud_index(self, cloud_index, cloud_index_name_map):
        # index -> cloud_name (lp_name) -> var object
        cloud_name = cloud_index_name_map.get(cloud_index)
        return self._lpname_to_var_map.get(cloud_name) if cloud_name else None

    def _make_solution(self, mdl, solution_handler):
        solver_name = self.docloud_solver_name
        # Store the results of solve in a solution object.
        docloud_obj = solution_handler.get_objective()
        docloud_values_by_idx, docloud_var_rcs = solution_handler.variable_results()
        # CPLEX index to name map
        # for those variables returned by CPLEX.
        # all other are assumed to be zero
        cloud_index_name_map = solution_handler.cplex_index_name_map()
        var_mapper = lambda idx: self._var_by_cloud_index(idx, cloud_index_name_map)
        # send an objective, a var-value dict and a string identifying the engine which solved.
        docloud_values_by_vars = {}
        keep_zeros = False
        count_nonmatching_cloud_vars = 0
        for cpx_idx, val in iteritems(docloud_values_by_idx):
            if keep_zeros or val:
                # first get the name from the cloud idx
                dvar = var_mapper(cpx_idx)
                if dvar:
                    docloud_values_by_vars[dvar] = val
                else:
                    cloud_name = cloud_index_name_map.get(cpx_idx)
                    if cloud_name and cloud_name.startswith("Rgc"):
                        # range variables
                        pass
                    else:
                        # one extra variable from docloud is OK
                        # it represents the constant term in objective
                        # more than one is an issue.
                        if count_nonmatching_cloud_vars:
                            mdl.info("Cannot find matching variable, cloud name is {0!s}", cloud_name)
                        count_nonmatching_cloud_vars += 1

        sol = SolveSolution.make_engine_solution(model=mdl,
                                                 obj=docloud_obj,
                                                 blended_obj_by_priority=[docloud_obj],
                                                 var_value_map=docloud_values_by_vars,
                                                 solved_by=solver_name,
                                                 solve_details=self._solve_details,
                                                 job_solve_status=self.get_solve_status())

        # attributes
        docloud_ct_duals, docloud_ct_slacks = solution_handler.constraint_results()

        ct_mapper = lambda idx: mdl.get_constraint_by_index(idx)
        sol.store_reduced_costs(docloud_var_rcs, mapper=var_mapper)
        sol.store_dual_values(docloud_ct_duals, mapper=ct_mapper)
        sol.store_slack_values(docloud_ct_slacks, mapper=ct_mapper)
        return sol

    def _make_relaxed_solution(self, mdl, solution_handler, infeas_handler):
        if infeas_handler is not None:
            infeasibilities = self._read_infeasibilities(mdl, infeas_handler)
        else:
            infeasibilities = self._compute_infeasibilities_from_slacks(mdl, solution_handler)

        sol = self._make_solution(mdl, solution_handler)
        sol.store_infeasibilities(infeasibilities)
        return sol

    def _read_infeasibilities(self, mdl, infeas_handler):
        # INTERNAL
        raw_infeasibilities = infeas_handler.get_infeasibilities()
        assert len(raw_infeasibilities) > 0
        infeasibilities = {}
        for ct_index, raw_infeas in iteritems(raw_infeasibilities):
            ct = mdl.get_constraint_by_index(ct_index)  # no exception: returns None if not found.
            if ct is not None:
                if raw_infeas != 0:
                    infeasibilities[ct] = raw_infeas
            else:
                self._unexpected_cloud_constraint_index(mdl, ct_index)
        return infeasibilities

    def _compute_infeasibilities_from_slacks(self, mdl, solution_handler):
        # INTERNAL - temporary
        raw_slacks = solution_handler.constraint_slacks()
        # from slack to infeasibility
        infeasibilities = {}
        for ct_index, raw_slack in iteritems(raw_slacks):
            ct = mdl.get_constraint_by_index(ct_index)  # no exception: returns None if not found.
            if ct is not None:
                infeas = ct.compute_infeasibility(raw_slack)
                if 0 != infeas:
                    infeasibilities[ct] = infeas
            else:
                # no constraint with this index, should not happen...
                self._unexpected_cloud_constraint_index(mdl, ct_index)

        return infeasibilities

    def _unexpected_cloud_constraint_index(self, mdl, ct_index):
        mdl.warning('unexpected index from cplex cloud: {}', ct_index)

    def get_solve_attribute(self, attr, index_seq):
        return {}

    def get_all_reduced_costs(self, mdl):
        return {}

    def get_solve_status(self):
        return self._connector.get_solve_status()

    def get_solve_details(self):
        return self._solve_details

    @classmethod
    def demangle_metric_name(cls, mname):
        from docplex.mp.constants import QualityMetric
        # find last occurence of '.' then identify the enum from the last part
        dotpos = mname.rfind('.')
        assert dotpos >= 0
        return QualityMetric.parse(mname[dotpos + 1:], raise_on_error=False)

    def _compute_quality_metrics(self, json_details):
        qms = {}
        if json_details:
            for qk, qv in iteritems(json_details):
                if 'quality.double' in qk:
                    qm = self.demangle_metric_name(qk)
                    if qm:
                        qms[qm.key] = float(qv)
                elif 'quality.int' in qk:
                    qm = self.demangle_metric_name(qk)
                    if qm:
                        iqv = int(qv)
                        if iqv >= 0:
                            qms[qm.int_key] = iqv
        return qms



# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2015, 2016, 2017, 2018
# --------------------------------------------------------------------------
# Author: Olivier OUDOT, IBM Analytics, France Lab, Sophia-Antipolis

"""
Parser converting a CPO file to internal model representation.
"""

from docplex.cp.cpo.cpo_tokenizer import *
from docplex.cp.expression import *
from docplex.cp.expression import _create_operation
from docplex.cp.function import *
from docplex.cp.catalog import *
from docplex.cp.solution import *
from docplex.cp.model import CpoModel
import docplex.cp.config as config
import math
import traceback


###############################################################################
## Constants
###############################################################################

# Minimum CPO format version number
MIN_CPO_VERSION_NUMBER = "12.6.0.0"

# Maximum CPO format version number
MAX_CPO_VERSION_NUMBER = "12.10.0.0"

# Map of all operators. Key is operator, value is list of corresponding operation descriptors
_ALL_OPERATORS = {}

# Map of all operations. Key is CPO operation name, value is list corresponding operation descriptor
_ALL_OPERATIONS = {}

# Initialization code
for op in ALL_OPERATIONS:
    if op.priority >= 0:
        _ALL_OPERATORS[op.keyword] = op
    _ALL_OPERATIONS[op.cpo_name] = op
_ALL_OPERATIONS['alldiff'] = Oper_all_diff
_ALL_OPERATIONS['allDiff'] = Oper_all_diff

# Known identifiers
_KNOWN_IDENTIFIERS = {"intmax": INT_MAX, "intmin": INT_MIN,
                      "inf": INFINITY,
                      "intervalmax": INTERVAL_MAX, "intervalmin": INTERVAL_MIN,
                      "no": False}

# Map of array types for each CPO name
_ARRAY_TYPES = {'intArray': Type_IntArray, 'floatArray': Type_FloatArray,
                'boolExprArray' : Type_BoolExprArray, 'intExprArray': Type_IntExprArray, 'floatExprArray': Type_FloatExprArray,
                'intervalVarArray': Type_IntervalVarArray, 'sequenceVarArray': Type_SequenceVarArray,
                'intValueSelectorArray': Type_IntValueSelectorArray, 'intVarSelectorArray': Type_IntVarSelectorArray,
                'tupleSet' : Type_TupleSet,
                '_cumulAtomArray' : Type_CumulAtomArray}

# Fake operator '..' to read intervals properly
_OPER_INTERVAL = CpoOperation("_interval", "_interval", "..", 9, (CpoSignature(Type_IntExprArray, (Type_IntExpr, Type_IntExpr)),))
_ALL_OPERATORS[".."] = _OPER_INTERVAL


###############################################################################
## Public classes
###############################################################################

class CpoParserException(CpoException):
    """ The base class for exceptions raised by the CPO parser
    """
    def __init__(self, msg):
        """ Create a new exception
        Args:
            msg: Error message
        """
        super(CpoParserException, self).__init__(msg)


class CpoUnsupportedFormatVersionException(CpoParserException):
    """ Exception raised when a version mismatch is detected
    """
    def __init__(self, msg):
        """ Create a new exception
        Args:
            msg: Error message
        """
        super(CpoUnsupportedFormatVersionException, self).__init__(msg)



class CpoParser(object):
    """ Reader of CPO file format """
    __slots__ = ('model',          # Read model
                 'expr_dict',      # Dictionary of expressions. Key is name, value is CPO expression
                 'source_file',    # Source file
                 'tokenizer',      # Reading tokenizer
                 'token',          # Last read token
                 'pushtoken',      # Pushed token
                 'fun_handlers',   # Special function handlers
                 'current_loc',    # Current source location
                 )
    
    def __init__(self, mdl=None):
        """ Create a new CPO format parser

        Args:
            mdl:  Model to fill, None (default) to create a new one.
        """
        super(CpoParser, self).__init__()
        self.model = mdl if mdl is not None else CpoModel()
        self.expr_dict = {}
        self.source_file = None
        self.tokenizer = None
        self.token = None
        self.pushtoken = None
        self.current_loc = None

        # Do not store location information (would store parser instead of real lines)
        self.model.source_loc = False

        # Initialize special function handlers
        self.fun_handlers = {
            'intVar':            self._read_fun_intVar,
            '_intVar':           self._read_fun_intVar,
            'boolVar':           self._read_fun_boolVar,
            '_boolVar':          self._read_fun_boolVar,
            'floatVar':          self._read_fun_floatVar,
            '_floatVar':         self._read_fun_floatVar,
            'intervalVar':       self._read_fun_intervalVar,
            'sequenceVar':       self._read_fun_sequenceVar,
            'stateFunction':     self._read_fun_stateFunction,
            'stepFunction':      self._read_fun_stepFunction,
            'segmentedFunction': self._read_fun_segmentedFunction,
            'transitionMatrix':  self._read_fun_transitionMatrix,
        }

        # TODO: parse and include source information ?


    def get_model(self):
        """ Get the model that have been parsed

        Return:
            Model result of the parsing, object of class :class:`~docplex.cp.model.CpoModel`
        """
        return self.model


    def parse(self, cfile):
        """ Parse a CPO file.

        Args:
            cfile: CPO file to read
        Return:
            Model result of the parsing, object of class :class:`~docplex.cp.model.CpoModel`
        Raises:
            CpoParserException: Parsing exception
            CpoVersionMismatchException: Read CPO format is not in [MIN_CPO_VERSION_NUMBER .. MAX_CPO_VERSION_NUMBER]
        """
        # Store file name if first file
        self.source_file = cfile
        if self.model.source_file is None:
            self.model.source_file = cfile

        self.tokenizer = CpoTokenizer(file=cfile)
        self._read_statement_list()
        self.tokenizer.close()
        self.tokenizer = None

        return self.model


    def parse_string(self, str):
        """ Parse a string.

        Result of the parsing is added to the current result model.

        Args:
            str: String to parse
        Return:
            Model result of the parsing, object of class :class:`~docplex.cp.model.CpoModel`
        """
        self.tokenizer = CpoTokenizer(input=str)
        self._read_statement_list()
        self.tokenizer.close()
        self.tokenizer = None
        return self.model


    def _read_statement_list(self):
        """ Read a list of statements

        This functions reads all statements up to end of input.
        """
        try:
            while self._read_statement():
                pass
        except Exception as e:
            #if isinstance(e, CpoParserException):
            #    raise e
            if config.context.log_exceptions:
                traceback.print_exc()
            self._raise_exception(str(e))
        return True


    def _read_statement(self):
        """ Read a statement or a section

        This functions reads the first token and exits with current token that is
        the last of the statement.

        Returns:
            True if something has been read, False if end of input
        """
        tok1 = self._next_token()
        if tok1 is TOKEN_EOF:
            return False
        tok2 = self._next_token()

        # Check obsolete let and set forms
        if tok1 in (TOKEN_LET, TOKEN_SET) and tok2.type is TOKEN_TYPE_SYMBOL:
            tok1 = tok2
            tok2 = self._next_token()

        if tok1 is TOKEN_HASH:
            self._read_directive()

        elif tok2 is TOKEN_ASSIGN:
            expr = self._read_assignment(tok1.get_string())
            # Add expression to the model if it is a variable
            if expr.is_variable():
                self.model._add_with_loc(expr, self.current_loc)

        elif tok2 is TOKEN_SEMICOLON:
            # Get existing expression and re-add it to the model
            expr = self.expr_dict.get(tok1.get_string())
            if expr is None:
                self._raise_exception("Expression '{}' not found in the model".format(tok1.get_string()))
            self.model._add_with_loc(expr, self.current_loc)

        elif tok2 is TOKEN_COLON:
            expr = self._read_assignment(tok1.get_string())
            self.model._add_with_loc(expr, self.current_loc)

        elif tok2 is TOKEN_BRACE_OPEN:
            self._read_section(tok1.value)

        else:
            # Read expression
            self._push_token(tok1)
            expr = self._read_expression()
            #print("Expression read in statement: Type: " + str(expr.type) + ", val=" + str(expr))
            self.model._add_with_loc(expr, self.current_loc)
            self._check_token(self.token, TOKEN_SEMICOLON)

        return True


    def _read_directive(self):
        """ Read a directive
        """
        name = self.token.value
        if name == "line":
            self._read_directive_line()

        elif name == "include":
            self._read_directive_include()

        else:
            self._raise_exception("Unknown directive '" + name + "'")


    def _read_directive_line(self):
        """ Read a line directive
        """
        # # Skip line
        # self.tokenizer._skip_to_end_of_line()
        tok = self._next_token()
        # Check line off
        if (tok.type == TOKEN_TYPE_SYMBOL) and tok.value == "off":
            self.current_loc = None
            return
        # Get line number
        if tok.type != TOKEN_TYPE_INTEGER:
            self._raise_exception("Line number should be an integer")
        lnum = int(tok.value)
        # Get optional source file name, string on the same line
        cline = self.tokenizer.line_number
        tok = self._next_token()
        if cline != self.tokenizer.line_number:
            self._push_token(tok)
            fname = None if self.current_loc is None else self.current_loc[0]
        else:
            if tok.type != TOKEN_TYPE_STRING:
               self._raise_exception("File name should be a string")
            fname = tok.get_string()
        self.current_loc = (fname, lnum)


    def _read_directive_include(self):
        """ Read a include directive
        """
        # Get file name
        fname = self._check_token_string(self._next_token())
        if (os.path.dirname(fname) == "") and (self.source_file is not None):
            fname = os.path.dirname(self.source_file) + "/" + fname
        # Push current context
        old_ctx = (self.source_file, self.tokenizer, self.token)
        # Parse file
        self.parse(fname)
        # Restore context
        self.source_file, self.tokenizer, self.token = old_ctx


    def _read_assignment(self, name):
        """ Read an assignment

        Args:
            name:  Assignment name
        Returns:
            Named expression
        """
        # Read expression
        tok = self._next_token()
        expr = self._read_expression()
        # Build CPO expression if needed
        expr = build_cpo_expr(expr)
        # Add name to the expression
        if expr.name:
            # Create an alias if needed
            if name != expr.name:
                expr = CpoAlias(expr, name)
                self.model.add(expr)
        else:
            expr.set_name(name)

        # Add expression to expressions dictionary
        self.expr_dict[name] = expr
        self._check_token(self.token, TOKEN_SEMICOLON)
        return expr


    def _read_interval_var(self):
        """ Read a interval_var declaration
        Returns with token set to the last unexpected token
        Returns:
            CpoIntervalVar variable expression
        """
        res = interval_var()
        while True:
            # Read argument name
            tok = self.token
            if tok.type is not TOKEN_TYPE_SYMBOL:
                return res
            aname = tok.value
            self._next_token()
            if aname == "present":
                res.set_present()
            elif aname == "absent":
                res.set_absent()
            elif aname == "optional":
                res.set_optional()
            else:
                self._check_token(self.token, TOKEN_ASSIGN)
                self._next_token()
                if aname in ("start", "end", "length", "size"):
                    # Read interval
                    intv = self._read_expression()
                    if isinstance(intv, int):
                        intv = (intv, intv)
                    elif not isinstance(intv, (list, tuple)):
                        self._raise_exception("'start', 'end', 'length' or 'size' should be an integer or an interval")
                    setattr(res, aname, intv)
                elif aname == "intensity":
                    res.set_intensity(self._read_expression())
                elif aname == "granularity":
                    self._check_token_integer(self.token)
                    res.set_granularity(int(self.token.value))
                    self._next_token()
                else:
                    self._raise_exception("Unknown IntervalVar attribute argument '" + aname + "'")
            # Read comma
            tok = self.token
            if tok.value == ',':
                tok = self._next_token()


    def _read_expression(self):
        """ Read an expression

        First expression token is already read.
        Function exits with current token following the last expression token

        Returns:
            Expression that has been read
        """
        # Read first sub-expression
        expr = self._read_sub_expression()
        tok = self.token
        if tok.type is not TOKEN_TYPE_OPERATOR:
            return expr

        # Initialize elements stack
        stack = [expr]
        while tok.type is TOKEN_TYPE_OPERATOR:
            op = self._get_and_check_operator(tok)
            self._next_token()
            expr = self._read_sub_expression()
            tok = self.token

            # Reduce stack if possible
            while (len(stack) > 1) and op.priority >= stack[-2].priority:
                oexpr = stack.pop()
                oop = stack.pop()
                stack[-1] = self._create_operation_expression(oop, (stack[-1], oexpr))

            stack.append(op)
            stack.append(expr)

        # Build final expression
        expr = stack.pop()
        while stack:
            op = stack.pop()
            expr = self._create_operation_expression(op, (stack.pop(), expr))
        return expr


    def _read_sub_expression(self):
        """ Read a sub-expression

        First expression token is already read.
        Function exits with current token following the last expression token
        Return:
            Expression that has been read
        """

        tok = self.token
        toktyp = tok.type
        ntok = self._next_token()

        # Check int constant
        if toktyp is TOKEN_TYPE_INTEGER:
            return int(tok.value)
        
        # Check float constant
        if toktyp is TOKEN_TYPE_FLOAT:
            return float(tok.value)

        # Check unary operator
        if toktyp is TOKEN_TYPE_OPERATOR:
            # Retrieve operation descriptor
            op = self._get_and_check_operator(tok)
            # Read next expression
            expr = self._read_sub_expression()
            return self._create_operation_expression(op, (expr,))
        
        # Check symbol
        if toktyp is TOKEN_TYPE_SYMBOL:
            tokval = tok.value
            if ntok is TOKEN_PARENT_OPEN:
                self._next_token()
                # Check special function calls
                fun = self.fun_handlers.get(tokval)
                if fun:
                    return fun()

                # General function call, retrieve operation descriptor
                op = _ALL_OPERATIONS.get(tokval)
                if op is None:
                    self._raise_exception("Unknown operation '" + str(tok.value) + "'")
                args = self._read_expression_list_up_to_parent_close()
                return self._create_operation_expression(op, args)

            if ntok is TOKEN_HOOK_OPEN:
                # Read typed array
                self._next_token()
                expr = self._read_expression_list(TOKEN_HOOK_CLOSE)
                # Search type
                typ = _ARRAY_TYPES.get(tok.value)
                if typ is None:
                    self._raise_exception("Unknown array type '" + str(tok.value) + "'")
                # Check empty array
                if len(expr) == 0:
                    return CpoValue((), typ)
                # Compute best type if array not empty
                res = build_cpo_expr(expr)
                if not res.type.is_kind_of(typ):
                    res.type = typ
                return res

            # Check known identifier
            if tokval in _KNOWN_IDENTIFIERS:
                return _KNOWN_IDENTIFIERS[tok.value]

            # Token is an expression id
            return self._get_identifier_value(tok.get_string())
        
        # Check expression in parenthesis
        if tok is TOKEN_PARENT_OPEN:
            expr = self._read_expression_list_up_to_parent_close()
            if len(expr) == 1:
                return expr[0]
            return expr

        # Check array with no type
        if tok is TOKEN_HOOK_OPEN:
            expr = self._read_expression_list(TOKEN_HOOK_CLOSE)
            return expr
            
        # Check reference to a model expression or variable
        if toktyp is TOKEN_TYPE_STRING:
            return self._get_identifier_value(tok.get_string())
                
        # Unknown expression
        self._raise_exception("Invalid start of expression: '" + str(tok) + "'")


    def _read_fun_intVar(self):
        """ Read a function call to 'intVar'
        Returns: New expression
        """
        args = self._read_expression_list_up_to_parent_close()
        return CpoIntVar(tuple(args))


    def _read_fun_intervalVar(self):
        """ Read a function call to 'intervalVar'
        Returns: New expression
        """
        expr = self._read_interval_var()
        self._check_token(self.token, TOKEN_PARENT_CLOSE)
        self._next_token()
        return expr


    def _read_fun_boolVar(self):
        """ Read a function call to 'boolVar'
        Returns: New expression
        """
        args = self._read_expression_list_up_to_parent_close()
        # Should add some test here to verify the domain
        return CpoBoolVar(tuple(args))


    def _read_fun_floatVar(self):
        """ Read a function call to 'floatVar'
        Returns: New expression
        """
        args = self._read_expression_list_up_to_parent_close()
        if len(args) != 2:
            self._raise_exception("'_floatVar' should have 2 arguments")
        return CpoFloatVar(args[0], args[1])


    def _read_fun_sequenceVar(self):
        """ Read a function call to 'sequenceVar'
        Returns: New expression
        """
        args = self._read_expression_list_up_to_parent_close()
        if len(args) == 1:
            lvars = args[0]
            ltypes = None
        else:
            if len(args) != 2:
                self._raise_exception("'sequenceVar' should have 1 or 2 arguments")
            lvars = args[0]
            ltypes = args[1]
        return CpoSequenceVar(lvars, ltypes)


    def _read_fun_stateFunction(self):
        """ Read a function call to 'stateFunction'
        Returns: New expression
        """
        args = self._read_expression_list_up_to_parent_close()
        nbargs = len(args)
        if nbargs == 0:
            trmx = None
        elif nbargs == 1:
            trmx = args[0]
        else:
            self._raise_exception("'stateFunction' should have 0 or 1 argument")
        return CpoStateFunction(trmx)


    def _read_fun_stepFunction(self):
        """ Read a function call to 'stepFunction'
        Returns: New expression
        """
        args = self._read_expression_list_up_to_parent_close()
        return CpoStepFunction(args)


    def _read_fun_segmentedFunction(self):
        """ Read a function call to 'segmentedFunction'
        Returns: New expression
        """
        args = self._read_expression_list_up_to_parent_close()
        return CpoSegmentedFunction(args[0], args[1:])


    def _read_fun_transitionMatrix(self):
        """ Read a function call to 'transitionMatrix'
        Returns: New expression
        """
        # Check arguments list to support presolved version
        if self.token.value == "matrixSize":
            args = self._read_arguments_list_up_to_parent_close()
            for name, mtrx in args:
                if name == 'matrix':
                    break
        else:
            mtrx = self._read_expression_list_up_to_parent_close()
        slen = len(mtrx)
        size = int(math.sqrt(slen))
        if size * size != slen:
            raise CpoParserException("Length of transition matrix values should be a square")
        return CpoTransitionMatrix(values=(mtrx[i * size : (i+1) * size] for i in range(size)))


    def _read_expression_list_in_parenthesis(self):
        """ Read a list of expressions between parenthesis
        Opening parenthesis is read and checked by this method.
        When returning, current token is token after closing parenthesis.
        Args:
           etok: Expression list ending token string (for example ')' or ']')
        Returns:
            Array of expressions
        """
        self._check_token(self._next_token(), TOKEN_PARENT_OPEN)
        self._next_token()
        lxpr = []
        while self.token is not TOKEN_PARENT_CLOSE:
            lxpr.append(self._read_expression())
            if self.token is TOKEN_COMMA:
                self._next_token()
        self._next_token()
        return tuple(lxpr)


    def _read_expression_list(self, etok):
        """ Read a list of expressions
        This method supposes that the current token is just after list starting character.
        When returning, current token is token after end of list

        Args:
           etok: Expression list ending token string (for example ')' or ']')
        Returns:
            Array of expressions
        """
        lxpr = []
        while self.token is not etok:
            lxpr.append(self._read_expression())
            if self.token is TOKEN_COMMA:
                self._next_token()
        self._next_token()
        return tuple(lxpr)


    def _read_expression_list_up_to_parent_close(self):
        """ Read a list of expressions up to a closing parenthesis
        This method supposes that current token is first after opening parenthesis.
        When returning, current token is token after end of list

        Returns:
            Array of expressions
        """
        lxpr = []
        while self.token is not TOKEN_PARENT_CLOSE:
            lxpr.append(self._read_expression())
            if self.token is TOKEN_COMMA:
                self._next_token()
        self._next_token()
        return tuple(lxpr)


    def _read_arguments_list_up_to_parent_close(self):
        """ Read a list of arguments that are possibly named, up to ending closing parenthesis

        This method supposes that the current token is list start (for example '(' or '[').
        When returning, current token is next to list ending token

        Args:
           etok: Expression list ending token (for example ')' or ']')
        Returns:
            Array of couples (name, expression)
        """
        lxpr = []
        while self.token is not TOKEN_PARENT_CLOSE:
            if self.token.type is TOKEN_TYPE_SYMBOL:
                name = self.token
                if self._next_token() is TOKEN_ASSIGN:
                    self._next_token()
                else:
                    self._push_token(name)
                    name = None
            else:
                name = None
            lxpr.append((name, self._read_expression()))
            if self.token is TOKEN_COMMA:
                self._next_token()
        self._next_token()
        return lxpr


    def _read_section(self, name):
        """ Read a section.
        Current token is the opening brace
        Args:
            name:  Section name
        """
        if name == "parameters":
            self._read_section_parameters()
        elif name == "internals":
            self._read_section_internals()
        elif name == "search":
            self._read_section_search()
        elif name == "startingPoint":
            self._read_section_starting_point()
        elif name == "KPIs":
            self._read_section_kpis()
        elif name == "phases":
            self._read_section_phases()
        else:
            self._raise_exception("Unknown section '" + name + "'")


    def _read_section_parameters(self):
        """ Read a parameters section
        """
        params = CpoParameters()
        tok = self._next_token()
        while (tok is not TOKEN_EOF) and (tok is not TOKEN_BRACE_CLOSE):
            vname = self._check_token_string(tok)
            self._check_token(self._next_token(), TOKEN_ASSIGN)
            value = self._next_token()
            self._check_token(self._next_token(), TOKEN_SEMICOLON)
            params.set_attribute(vname, value.get_string())
            tok = self._next_token()
        if params:
            self.model.set_parameters(params)


    def _read_section_internals(self):
        """ Read a internals section
        """
        # Skip all until section end
        tok = self._next_token()
        while (tok is not TOKEN_EOF) and (tok is not TOKEN_BRACE_CLOSE):
            if self.token.value == "version":
                self._check_token(self._next_token(), TOKEN_PARENT_OPEN)
                vtok = self._next_token()
                if vtok.type is TOKEN_TYPE_VERSION:
                    ver = vtok.get_string()
                    self.model.set_format_version(ver)
                    if compare_natural(ver, MIN_CPO_VERSION_NUMBER) < 0:
                        raise CpoUnsupportedFormatVersionException("Can not parse a CPO file with version {}, lower than {}"
                                                                   .format(ver, MIN_CPO_VERSION_NUMBER))
                    if compare_natural(ver, MAX_CPO_VERSION_NUMBER) > 0:
                        raise CpoUnsupportedFormatVersionException("Can not parse a CPO file with version {}, greater than {}"
                                                                   .format(ver, MAX_CPO_VERSION_NUMBER))
                self._check_token(self._next_token(), TOKEN_PARENT_CLOSE)
            tok = self._next_token()


    def _read_section_phases(self):
        """ Read a phase section (old CPO versions, ignored)
        """
        # Skip all until section end
        tok = self._next_token()
        while (tok is not TOKEN_EOF) and (tok is not TOKEN_BRACE_CLOSE):
            tok = self._next_token()


    def _read_section_search(self):
        """ Read a search section
        """
        # Read statements up to end of section
        tok = self._next_token()
        while (tok is not TOKEN_EOF) and (tok is not TOKEN_BRACE_CLOSE):
            self._push_token(tok)
            self._read_statement()
            tok = self._next_token()


    def _read_section_starting_point(self):
        """ Read a starting point section
        """
        sp = CpoModelSolution()
        # Read statements up to end of section
        tok = self._next_token()
        while (tok is not TOKEN_EOF) and (tok is not TOKEN_BRACE_CLOSE):
            # Check directive (#line)
            if tok is TOKEN_HASH:
                self._next_token()
                self._read_directive()
                tok = self._next_token()
                continue

            # Check token is a string
            vname = self._check_token_string(tok)

            # Check experimental section "expressions"
            if vname == "expressions":
                # Read opening brace
                self._check_token(self._next_token(), TOKEN_BRACE_OPEN)
                # Read and ignore all up to next brace close
                tok = self._next_token()
                while (tok is not TOKEN_EOF) and (tok is not TOKEN_BRACE_CLOSE):
                    tok = self._next_token()
                self._next_token()
                continue

            # Get and check the variable that is concerned
            var = self.expr_dict.get(vname)
            if var is None:
                self._raise_exception("There is no variable named '{}' in this model".format(vname))
            self._check_token(self._next_token(), TOKEN_ASSIGN)
            tok = self._next_token()
            vsol = None

            # Process integer variable
            if var.type is Type_IntVar:
                if tok.value == "intVar":
                    tok = self._next_token()
                if tok is TOKEN_PARENT_OPEN:
                    self._next_token()
                # Read domain
                dom = self._read_expression()
                if self.token is TOKEN_PARENT_CLOSE:
                    self._next_token()
                # Add solution to starting point
                vsol = CpoIntVarSolution(var, (dom, ))

            # Process interval variable
            elif var.type is Type_IntervalVar:
                if tok.value == "absent":
                    vsol = CpoIntervalVarSolution(var, presence=False)
                    self._next_token()
                else:
                    if tok.value == "intervalVar":
                        tok = self._next_token()
                    if tok is TOKEN_PARENT_OPEN:
                        self._next_token()
                        ivar = self._read_interval_var()
                        self._check_token(self.token, TOKEN_PARENT_CLOSE)
                        self._next_token()
                    else:
                        ivar = self._read_interval_var()
                    #ivar.set_name(vname)
                    vsol = CpoIntervalVarSolution(var,
                                                  presence=True if ivar.is_present() else False if ivar.is_absent() else None,
                                                  start=ivar.get_start() if ivar.get_start() != DEFAULT_INTERVAL else None,
                                                  end=ivar.get_end() if ivar.get_end() != DEFAULT_INTERVAL else None,
                                                  size=ivar.get_size() if ivar.get_size() != DEFAULT_INTERVAL else None)

            # Process sequence variable (not public)
            elif var.type is Type_SequenceVar:
                # Read array of variables
                expr = self._read_expression()
                if not is_array(expr):
                    self._raise_exception("In section 'startingPoint', the solution of a sequence variable should be a list of interval variables")
                vsol = CpoSequenceVarSolution(var, expr)

            # Process state function
            elif var.type is Type_StateFunction:
                # Read state function value and ignore
                self._read_expression()

            # Process pulse function
            elif var.type is Type_CumulAtom:
                # Read pulse expression and ignore
                self._read_expression()

            else:
                self._raise_exception("The section 'startingPoint' should contain only integer and interval variables.")

            # Add variable solution to starting point
            if vsol is not None:
                sp.add_var_solution(vsol)

            # Read end of variable starting point
            self._check_token(self.token, TOKEN_SEMICOLON)
            tok = self._next_token()

        # Add starting point to the model
        self.model.set_starting_point(sp)


    def _read_section_kpis(self):
        """ Read a KPI section
        """
        # Check that format version is appropriate
        fver = self.model.get_format_version()
        if fver is not None and compare_natural(fver, '12.9') < 0:
            raise CpoUnsupportedFormatVersionException("Section KPIs is not supported in format version {}".format(fver))

        # Read statements up to end of section
        tok = self._next_token()
        while (tok is not TOKEN_EOF) and (tok is not TOKEN_BRACE_CLOSE):
            # Check directive (line)
            if tok is TOKEN_HASH:
                self._next_token()
                self._read_directive()
                tok = self._next_token()
                continue

            # Get KPI name
            kname = self._check_token_string(tok)
            if self._next_token() is TOKEN_SEMICOLON:
                # Verify KPI name exists as an expression
                expr = self.expr_dict.get(kname)
                if expr is None:
                    self._raise_exception("There is no expression named '{}' in this model".format(kname))
            else:
                # Read expression
                self._check_token(self.token, TOKEN_ASSIGN)
                self._next_token()
                expr = self._read_expression()
                self._check_token(self.token, TOKEN_SEMICOLON)
            tok = self._next_token()
            # Add KPI
            self.model.add_kpi(expr, kname)


    def _check_token(self, tok, etok):
        """ Check that a read token is a given one an raise an exception if not
        Args:
            tok: Read token
            etok: Expected token
        """
        if tok is not etok:
            self._raise_unexpected_token(etok, tok)


    def _get_and_check_operator(self, tok):
        """ Get an operator descriptor and raise an exception if not found
        Args:
            tok:  Operator token
        Returns:
            List of Operation descriptor for this keyword
        Raises:
            CpoException if operator does not exists
        """
        op = _ALL_OPERATORS.get(tok.value)
        if op is None:
            self._raise_exception("Unknown operator '" + str(tok.value) + "'")
        return op


    def _check_token_string(self, tok):
        """ Check that a token is a string and raise an exception if not
        Args:
            tok: Token
        Returns:
            String value of the token            
        """
        if tok.type is TOKEN_TYPE_SYMBOL:
            return tok.value
        if tok.type is TOKEN_TYPE_STRING:
            return tok.get_string()
        self._raise_exception("String expected")


    def _check_token_integer(self, tok):
        """ Check that a token is an integer and raise an exception if not
        Args:
            tok: Token
        Returns:
            integer value of the token
        """
        if tok.type is TOKEN_TYPE_INTEGER:
            return int(tok.value)
        if tok.value in _KNOWN_IDENTIFIERS:
            return _KNOWN_IDENTIFIERS[tok.value]
        self._raise_exception("Integer expected instead of '" + tok.value + "'")


    def _get_identifier_value(self, eid):
        """ Get an expression associated to an identifier
        Args:
            eid:  Expression identifier
        Returns:
            Expression corresponding to this identifier
        """
        try:
            return self.expr_dict[eid]
        except:
            self._raise_exception("Unknown identifier '" + str(eid) + "'")


    def _create_operation_expression(self, op, args):
        """ Create a model operation

        Args:
            op:    Operation descriptor
            args:  Operation arguments
        Returns:
            Model expression
        Raises:
            Cpo exception if error
        """
        # Check interval operator
        if op is _OPER_INTERVAL:
            return tuple(args)
        # Check unary operations on constant value
        if (op is Oper_minus) and (len(args) == 1) and is_number(args[0]):
            return -args[0]
        if (op is Oper_plus) and (len(args) == 1) and is_number(args[0]):
            return args[0]

        try:
            return _create_operation(op, args)
        except Exception as e:
            lastex = Exception("No valid operation found for {}: {}".format(op.cpo_name, e))
            self._raise_exception(str(lastex))


    def _raise_unexpected_token(self, expect=None, tok=None):
        """ Raise a "Unexpected token" exception
        Args:
            tok:  Unexpected token
        """
        if tok is None: 
            tok = self.token
        if expect is None:
            self._raise_exception("Unexpected token '" + str(tok) + "'")
        self._raise_exception("Read '" + str(tok) + "' instead of expected '" + str(expect) + "'")


    def _raise_exception(self, msg):
        """ Raise a Parsing exception
        Args:
            msg:  Exception message
        """
        raise CpoParserException(self.tokenizer.build_error_string(msg))


    def _next_token(self):
        """ Read next token
        Returns:
            Next read token, None if end of input
        """
        # Check if a token has been pushed
        if self.pushtoken is not None:
            tok = self.pushtoken
            self.pushtoken = None
        else:
            tok = self.tokenizer.next_token()
        self.token = tok
        #print("Tok='" + str(tok) + "'")
        return tok


    def _push_token(self, tok):
        """ Push current token
        Args:
            tok: New current token 
        """
        self.pushtoken = self.token
        self.token = tok
        



# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2015, 2016
# --------------------------------------------------------------------------

# gendoc: ignore

from six import itervalues, iteritems
from docplex.mp.compat23 import izip

from docplex.mp.xcounter import update_dict_from_item_value

from docplex.mp.utils import is_number, is_iterable, is_iterator, is_pandas_series, \
    is_numpy_ndarray, is_pandas_dataframe, is_numpy_matrix, is_ordered_sequence
from docplex.mp.linear import Var, MonomialExpr, AbstractLinearExpr, LinearExpr, ZeroExpr
from docplex.mp.functional import _FunctionalExpr
from docplex.mp.operand import LinearOperand
from docplex.mp.quad import QuadExpr, VarPair


class ModelAggregator(object):
    # what type to use for merging dicts

    def __init__(self, linear_factory, quad_factory):
        self._linear_factory = linear_factory
        self._checker = linear_factory._checker
        self._quad_factory = quad_factory
        self._model = linear_factory._model
        self._generate_transients = True

    @property
    def counter_type(self):
        return self._linear_factory.term_dict_type

    def new_zero_expr(self):
        return ZeroExpr(model=self._model)

    def _to_expr(self, qcc, lcc=None, constant=0):
        # no need to sort here, sort is done by str() on the fly.
        if qcc:
            linear_expr = LinearExpr(self._model, e=lcc, constant=constant, safe=True)
            quad = self._quad_factory.new_quad(quads=qcc, linexpr=linear_expr, safe=True)
            quad._transient = self._generate_transients
            return quad
        elif lcc or constant:
            linear_expr = LinearExpr(self._model, e=lcc, constant=constant, safe=True)
            linear_expr._transient = self._generate_transients
            return linear_expr
        else:
            return self.new_zero_expr()

    def scal_prod(self, terms, coefs=1.0):
        # Testing anumpy array for its logical value will not work:
        # ValueError: The truth value of an array with more than one element is ambiguous. Use a.any() or a.all()
        # we would have to trap the test for ValueError then call any()
        #
        if is_iterable(coefs):
            pass  # ok
        elif is_number(coefs):
            if 0 == coefs:
                return self.new_zero_expr()
            else:
                sum_expr = self.sum(terms)
                return sum_expr * coefs
        else:
            self._model.fatal("scal_prod expects iterable or number, {0!s} was passed", coefs)

        # model has checked terms is an ordered sequence
        return self._scal_prod(terms, coefs)

    def _scal_prod(self, terms, coefs):
        # INTERNAL
        checker = self._checker
        total_num = 0
        lcc = self.counter_type()
        qcc = None

        number_validation_fn = checker.get_number_validation_fn()

        for item, coef in izip(terms, coefs):
            if not coef:
                continue

            safe_coef = number_validation_fn(coef) if number_validation_fn else coef
            if isinstance(item, Var):
                update_dict_from_item_value(lcc, item, safe_coef)

            elif isinstance(item, AbstractLinearExpr):
                total_num += safe_coef * item.get_constant()
                for lv, lk in item.iter_terms():
                    update_dict_from_item_value(lcc, lv, lk * safe_coef)

            elif isinstance(item, QuadExpr):
                if qcc is None:
                    qcc = self.counter_type()
                for qv, qk in item.iter_quads():
                    update_dict_from_item_value(qcc, qv, qk * safe_coef)
                qlin = item.get_linear_part()
                for v, k in qlin.iter_terms():
                    update_dict_from_item_value(lcc, v, k * safe_coef)

                total_num += safe_coef * qlin.constant

            # --- try conversion ---
            else:
                try:
                    e = item.to_linear_expr()
                    total_num += e.get_constant()
                    for dv, k, in e.iter_terms():
                        update_dict_from_item_value(lcc, dv, k * safe_coef)
                except AttributeError:
                    self._model.fatal("scal_prod accepts variables, expressions, numbers, not: {0!s}", item)

        return self._to_expr(qcc, lcc, total_num)

    def _scal_prod_f(self, dvars, coef_fn, assume_alldifferent):
        if isinstance(dvars, dict) and hasattr(dvars, 'items'):
            var_key_iter = iteritems
        elif is_ordered_sequence(dvars):
            var_key_iter = enumerate
        else:
            var_key_iter = None
            self._model.fatal('Model.dotf expects either a dictionary or an ordered sequence of variables, an instance of {0} was passed',
                              type(dvars))

        if assume_alldifferent:
            return self._scal_prod_f_alldifferent(dvars, coef_fn, var_key_iter)
        else:
            return self._scal_prod_f_gen(dvars, coef_fn, var_key_iter=var_key_iter)

    def _scal_prod_f_gen(self, dvars, coef_fn, var_key_iter):
        # var_map is a dictionary of variables.
        # coef_fn is a function accepting dictionary keys
        lcc_type = self.counter_type
        lcc = lcc_type()
        number_validation_fn = self._checker.get_number_validation_fn()
        if number_validation_fn:
            for k, dvar in var_key_iter(dvars):
                fcoeff = coef_fn(k)
                safe_coeff = number_validation_fn(fcoeff)
                if safe_coeff:
                    update_dict_from_item_value(lcc, dvar, safe_coeff)
        else:
            for k, dvar in var_key_iter(dvars):
                fcoeff = coef_fn(k)
                if fcoeff:
                    update_dict_from_item_value(lcc, dvar, fcoeff)

        return self._to_expr(qcc=None, lcc=lcc)

    def _scal_prod_f_alldifferent(self, dvars, coef_fn, var_key_iter):
        # var_map is a dictionary of variables.
        # coef_fn is a function accepting dictionary keys
        lcc_type = self.counter_type
        lcc = lcc_type()
        lcc_setitem = lcc_type.__setitem__
        number_validation_fn = self._checker.get_number_validation_fn()
        if number_validation_fn:
            for k, dvar in var_key_iter(dvars):
                fcoeff = coef_fn(k)
                safe_coeff = number_validation_fn(fcoeff)
                if safe_coeff:
                    lcc_setitem(lcc, dvar, safe_coeff)
        else:
            for k, dvar in var_key_iter(dvars):
                fcoeff = coef_fn(k)
                if fcoeff:
                    lcc_setitem(lcc, dvar, fcoeff)

        return self._to_expr(qcc=None, lcc=lcc)

    def sum(self, sum_args):
        if is_iterator(sum_args):
            return self._sum_with_iter(sum_args)

        elif is_numpy_ndarray(sum_args):
            return self._sum_with_iter(sum_args.flat)
        elif is_pandas_series(sum_args):
            return self.sum(sum_args.values)
        elif isinstance(sum_args, dict):
            # handle dict: sum all values
            return self._sum_with_iter(itervalues(sum_args))
        elif is_iterable(sum_args):
            return self._sum_with_seq(sum_args)

        elif is_number(sum_args):
            return sum_args
        else:
            # TODO: use linear operand here
            return self._linear_factory._to_linear_operand(sum_args)

    def _sum_with_iter(self, args):
        sum_of_nums = 0
        lcc = self.counter_type()
        checker = self._checker
        qcc = None
        number_validation_fn = checker.get_number_validation_fn()
        for item in args:
            if isinstance(item, LinearOperand):
                for lv, lk in item.iter_terms():
                    update_dict_from_item_value(lcc, lv, lk)
                itc = item.get_constant()
                if itc:
                    sum_of_nums += itc
            elif is_number(item):
                sum_of_nums += number_validation_fn(item) if number_validation_fn else item
            elif isinstance(item, QuadExpr):
                for lv, lk in item.linear_part.iter_terms():
                    update_dict_from_item_value(lcc, lv, lk)
                if qcc is None:
                    qcc = self.counter_type()
                for qvp, qk in item.iter_quads():
                    update_dict_from_item_value(qcc, qvp, qk)
                sum_of_nums += item.get_constant()

            else:
                try:
                    expr = item.to_linear_expr()
                    sum_of_nums += expr.get_constant()
                    for dv, k in expr.iter_terms():
                        update_dict_from_item_value(lcc, dv, k)
                except AttributeError:
                    self._model.fatal("Model.sum() expects numbers/variables/expressions, got: {0!s}", item)

        return self._to_expr(qcc, lcc, sum_of_nums)

    def _sum_vars(self, dvars):
        if is_numpy_ndarray(dvars):
            return self._sum_vars(dvars.flat)
        elif is_pandas_series(dvars):
            return self.sum(dvars.values)
        elif isinstance(dvars, dict):
            # handle dict: sum all values
            return self._sum_vars(itervalues(dvars))
        elif is_iterable(dvars):
            checked_dvars = self._checker.typecheck_var_seq(dvars, caller='Model.sumvars()')
            sumvars_terms = self._varlist_to_terms(checked_dvars)
            return self._to_expr(qcc=None, lcc=sumvars_terms)
        else:
            self._model.fatal('Model.sumvars() expects an iterable returning variables, {0!r} was passed',
                              dvars)

    def _sum_vars_all_different(self, dvars):
        lcc = self._linear_factory._new_term_dict()
        setitem_fn = lcc.__setitem__

        for v in dvars:
            setitem_fn(v, 1)
        return self._to_expr(qcc=None, lcc=lcc)

    def _varlist_to_terms(self, var_list):
        # INTERNAL: converts a sum of vars to a dict, sorting if needed.
        linear_term_dict_type = self._linear_factory.term_dict_type
        try:
            assume_no_dups = len(var_list) == len(set(var_list))
        except TypeError:
            assume_no_dups = False

        if assume_no_dups:
            varsum_terms = linear_term_dict_type()
            linear_terms_setitem = linear_term_dict_type.__setitem__
            for v in var_list:
                linear_terms_setitem(varsum_terms, v, 1)
        else:
            # there might be repeated variables.
            varsum_terms = linear_term_dict_type()
            for v in var_list:
                update_dict_from_item_value(varsum_terms, v, 1)
        return varsum_terms

    def _sum_with_seq(self, sum_args):
        for z in sum_args:
            if not isinstance(z, Var):
                x_seq_all_variables = False
                break
        else:
            x_seq_all_variables = True

        if x_seq_all_variables:
            return self._sum_vars(sum_args)
        else:
            return self._sum_with_iter(args=sum_args)

    def _sumsq(self, args):
        accumulated_ct = 0
        number_validation_fn = self._checker.get_number_validation_fn()
        qcc = self._quad_factory.term_dict_type()
        lcc = self._linear_factory.term_dict_type()

        for item in args:
            if isinstance(item, Var):
                update_dict_from_item_value(qcc, VarPair(item, item), 1)
            elif isinstance(item, MonomialExpr):
                mcoef = item._coef
                # noinspection PyPep8
                mvar = item._dvar
                update_dict_from_item_value(qcc, VarPair(mvar, mvar), mcoef ** 2)

            elif isinstance(item, LinearExpr):
                cst = item.get_constant()
                accumulated_ct += cst ** 2
                for lv1, lk1 in item.iter_terms():
                    for lv2, lk2 in item.iter_terms():
                        if lv1 is lv2:
                            update_dict_from_item_value(qcc, VarPair(lv1, lv1), lk1 * lk1)
                        elif lv1._index < lv2._index:
                            update_dict_from_item_value(qcc, VarPair(lv1, lv2), 2 * lk1 * lk2)
                        else:
                            pass

                    if cst:
                        update_dict_from_item_value(lcc, lv1, 2 * cst * lk1)
            elif isinstance(item, _FunctionalExpr):
                fvar = item.functional_var
                update_dict_from_item_value(qcc, VarPair(fvar), 1)

            elif isinstance(item, ZeroExpr):
                pass

            elif is_number(item):
                safe_item = number_validation_fn(item) if number_validation_fn else item
                accumulated_ct += safe_item ** 2

            else:
                self._model.fatal("Model.sumsq() expects numbers/variables/linear expressions, got: {0!s}", item)

        return self._to_expr(qcc, lcc, constant=accumulated_ct)

    def sumsq(self, sum_args):
        if is_iterable(sum_args):
            if is_iterator(sum_args):
                return self._sumsq(sum_args)
            elif isinstance(sum_args, dict):
                return self._sumsq(sum_args.values())
            elif is_numpy_ndarray(sum_args):
                return self._sumsq(sum_args.flat)
            elif is_pandas_series(sum_args):
                return self._sumsq(sum_args.values)

            else:
                return self._sumsq(sum_args)
        elif is_number(sum_args):
            return sum_args ** 2
        else:
            self._model.fatal("Model.sumsq() expects number/iterable/expression, got: {0!s}", sum_args)

    # --- matrix constraint
    @staticmethod
    def generate_df_rows(df):
        for row in df.itertuples(index=False):
            yield row

    @staticmethod
    def generate_np_matrix_rows(npm):
        for r in npm:
            yield r.tolist()[0]

    def _sparse_make_exprs(self, sp_mat, dvars, nb_exprs):
        lfactory = self._linear_factory
        exprs = [lfactory.linear_expr() for _ in range(nb_exprs)]
        coo_mat = sp_mat.tocoo()
        for coef, row, col in izip(coo_mat.data, coo_mat.row, coo_mat.col):
            exprs[row]._add_term(dvars[col], coef)
        return exprs

    def _sparse_matrix_constraints(self, sp_coef_mat, svars, srhs, op):
        range_cts = range(len(srhs))
        lfactory = self._linear_factory
        exprs = self._sparse_make_exprs(sp_coef_mat, svars, len(srhs))
        cts = [lfactory._new_binary_constraint(exprs[r], sense=op, rhs=srhs[r]) for r in range_cts]
        return cts

    @classmethod
    def generate_rows(cls, coef_mat):
        if is_pandas_dataframe(coef_mat):
            row_gen = cls.generate_df_rows(coef_mat)
        elif is_numpy_matrix(coef_mat):
            row_gen = cls.generate_np_matrix_rows(coef_mat)
        else:
            row_gen = iter(coef_mat)
        return row_gen

    def _matrix_constraints(self, coef_mat, svars, srhs, op):
        row_gen = self.generate_rows(coef_mat)
        lfactory = self._linear_factory

        return [lfactory._new_binary_constraint(lhs=self._scal_prod(svars, row), sense=op, rhs=rhs)
                for row, rhs in izip(row_gen, srhs)]

    def _matrix_ranges(self, coef_mat, svars, lbs, ubs):
        row_gen = self.generate_rows(coef_mat)
        lfactory = self._linear_factory

        return [lfactory.new_range_constraint(expr=self._scal_prod(svars, row), lb=lb, ub=ub)
                for row, lb, ub in izip(row_gen, lbs, ubs)]

    def _sparse_matrix_ranges(self, sp_coef_mat, svars, lbs, ubs):
        assert len(lbs) == len(ubs)
        range_ranges = range(len(lbs))
        lfactory = self._linear_factory
        exprs = self._sparse_make_exprs(sp_coef_mat, svars, nb_exprs=len(lbs))
        rgs = [lfactory.new_range_constraint(lbs[r], exprs[r], ubs[r], check_feasible=False) for r in range_ranges]
        return rgs



# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2019
# --------------------------------------------------------------------------
# Author: Olivier OUDOT, IBM Analytics, France Lab, Sophia-Antipolis

"""
This module defines the class :class:`CpoCallback` that allows to retrieve events
that are sent by the CP Optimizer solver engine when running.

Any number of callbacks can be added to a solver using the method :meth:`docplex.cp.solver.solver.CpoSolver.add_callback`.
Callbacks can also be added on the model object using :meth:`docplex.cp.model.CpoModel.add_callback`

*New in version 2.10.*

Detailed description
--------------------
"""


#==============================================================================
#  Constants
#==============================================================================

# Call back events
EVENT_START_SOLVE               = "StartSolve"
EVENT_END_SOLVE                 = "EndSolve"
EVENT_START_EXTRACTION          = "StartExtraction"
EVENT_END_EXTRACTION            = "EndExtraction"
EVENT_START_INITIAL_PROPAGATION = "StartInitialPropagation"
EVENT_END_INITIAL_PROPAGATION   = "EndInitialPropagation"
EVENT_START_SEARCH              = "StartSearch"
EVENT_END_SEARCH                = "EndSearch"
EVENT_PERIODIC                  = "Periodic"
EVENT_OBJ_BOUND                 = "ObjBound"
EVENT_SOLUTION                  = "Solution"
EVENT_PROOF                     = "Proof"
EVENT_DESTRUCTION               = "Destruction"

ALL_CALLBACK_EVENTS = (EVENT_START_SOLVE, EVENT_END_SOLVE, EVENT_START_EXTRACTION, EVENT_END_EXTRACTION,
                       EVENT_START_INITIAL_PROPAGATION, EVENT_END_INITIAL_PROPAGATION, EVENT_START_SEARCH, EVENT_END_SEARCH,
                       EVENT_PERIODIC, EVENT_OBJ_BOUND, EVENT_SOLUTION, EVENT_DESTRUCTION,)


#==============================================================================
#  CPO callback class
#==============================================================================

class CpoCallback(object):
    """ CPO callback allows to be warned directly by the solver engine about different solving steps.

    This class is an 'abstract' class that must be extended by actual listener implementation.
    All method of this class are empty.

    *New in version 2.10.*
    """

    def invoke(self, solver, event, sres):
        """ Notify the callback about a solver event.

        This method is called every time an event is notified by the CPO solver.
        Associated to the event, the solver information is provided as a an object of class
        class:`~docplex.cp.solution.CpoSolveResult` that is instantiated with information available at this step.

        Args:
            solver: Originator CPO solver (object of class :class:`~docplex.cp.solver.solver.CpoSolver`)
            event:  Event id, string with value in ALL_CALLBACK_EVENTS
            sres:   Solver data, object of class :class:`~docplex.cp.solution.CpoSolveResult`
        """
        pass





# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2015, 2016
# --------------------------------------------------------------------------

# gendoc: ignore
from __future__ import print_function

from docplex.mp.compat23 import StringIO
from docplex.mp.constants import CplexScope
from docplex.mp.utils import OutputStreamAdapter

import sys

from six import iteritems


class ModelAnnotationPrinter(object):
    # header contains the final newline
    mst_header = """<?xml version = "1.0" standalone="yes"?>
"""
    anno_extension = ".ann"

    annotations_start_tag = "<CPLEXAnnotations>\n"
    annotations_end_tag   = "</CPLEXAnnotations>\n"
    benders_start_tag     = " <CPLEXAnnotation name='cpxBendersPartition' type='long' default='0'>\n"
    benders_end_tag       = " </CPLEXAnnotation>\n"

    # map to cplex annotation object types
    cplex_anno_objtype_map = {CplexScope.VAR_SCOPE: 1,
                              CplexScope.LINEAR_CT_SCOPE: 2,
                              CplexScope.SOS_SCOPE: 3,
                              CplexScope.IND_CT_SCOPE: 4,
                              CplexScope.QUAD_CT_SCOPE: 5}

    @staticmethod
    def print_signature(out):
        from docplex.version import docplex_version_string
        osa = OutputStreamAdapter(out)
        osa.write("<?xml version='1.0' encoding='utf-8'?>\n")
        osa.write("<!-- This file has been generated by DOcplex version {}  -->\n".format(docplex_version_string))

    @classmethod
    def _anno_name(cls, cpxscope, obj_name, obj_index):
        if obj_name is not None:
            return obj_name
        else:
            return '%s%d' % (cpxscope.prefix, obj_index)

    @classmethod
    def print(cls, out, model):
        osa = OutputStreamAdapter(out)

        cls.print_signature(out)
        # <CPLEXSolution version="1.0">
        osa.write(cls.annotations_start_tag)

        osa.write(cls.benders_start_tag)
        for cplex_scope, annotated in iteritems(model.get_annotations_by_scope()):
            anno_objtype = cls.cplex_anno_objtype_map.get(cplex_scope)
            if anno_objtype is not None and annotated:
                osa.write("  <object type='{0:d}'>\n".format(anno_objtype))
                for obj, benders_value in annotated:                   # FIXME: what if no name??
                    obj_index = obj.get_index()  # not checked
                    anno_name = cls._anno_name(cplex_scope, obj.get_name(), obj_index)
                    if obj_index >= 0 :
                        # ignore objects not added to the model
                        osa.write("   <anno name='{0}' index='{1:d}' value='{2:d}'/>\n".format(anno_name, obj_index, benders_value))
                osa.write('  </object>\n')
        #  </CPLEXAnnotations>
        osa.write(cls.benders_end_tag)
        osa.write(cls.annotations_end_tag)
        osa.write('\n')

    @classmethod
    def print_to_stream(cls, mdl, out, extension=anno_extension):
        if out is None:
            # prints on standard output
            cls.print(sys.stdout, mdl)
        elif isinstance(out, str):
            # a string is interpreted as a path name
            path = out if out.endswith(extension) else out + extension
            with open(path, "w") as of:
                cls.print_to_stream(mdl, of)
                # print("* file: %s overwritten" % path)
        else:
            try:
                cls.print(out, mdl)

            except AttributeError:  # pragma: no cover
                pass  # pragma: no cover
                # stringio will raise an attribute error here, due to with
                # print("Cannot use this an output: %s" % str(out))

    @classmethod
    def print_to_string(cls, mdl):
        oss = StringIO()
        cls.print_to_stream(mdl, out=oss)
        return oss.getvalue()



#!/usr/bin/python
# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2017
# ---------------------------------------------------------------------------
#
# gendoc: ignore
from docplex.mp.compat23 import izip
from docplex.mp.cplex_engine import CplexEngine


class ModelCallbackMixin(object):
    """
    This mixin class is intended as a bridge between DOcplex expression and constraints
    and CPLEX callback API.
    It is not intended to be instantiated directly, but to be inherited from in custom callbacks
    , jointly with a CPLEX callback type.

    For example, to define a custom BranchCallback in Docplex, define a new class which
    inherits both from ModelCallbackMixin and the legacy callback class BranchCallback.

    Note:
        - `ModelCallbackMixin` should be first in inheritance order,
        - the constructor of the custom callback class must take an `env` parameter to comply
            with the CPLEX API
        - the constructor of the custom callback must call two __init__() methods:
            - one for the cplex callback class, taking an `env` parameter
            - one for the mixin class.

    Example:

        class MyBranch(ModelCallbackMixin, cplex.callbacks.BranchCallback):

            def __init__(self, env):
                cplex.callbacks.BranchCallback.__init__(self, env)
                ModelCallbackMixin.__init__(self)

    A custom callback must be registered with a `Model` class using Model.register_callback; this method
    assumes the custom callback has a model setter property to connect the model to the callback.


        See Also:
            :func:`docplex.mp.model.Model.register_callback`
    """
    def __init__(self):
        self._model = None

    @property
    def model(self):
        """ This property is used to get the model associated with the mixin.

        An exception is raised if no model has been associated with the mixin.

        :return: an instance of `docplex.mp.Model`
        """
        if not self._model:
            raise ValueError('No model has been attached to the callback.')  # pragma: no cover
        return self._model

    def index_to_var(self, var_index):
        """ This method converts a variable index to a Var object.

        A model must have been associated withthe mixin, otherwise an error is raised.

        :param var_index: A valid variable index, that is a positive integer.

        :return: A Docplex variable with this index, or None.
        """
        assert var_index >= 0
        dv = self.model.get_var_by_index(var_index)
        return dv

    @staticmethod
    def linear_ct_to_cplex(linear_ct):
        cpx_lhs = CplexEngine.linear_ct_to_cplex(linear_ct=linear_ct)
        cpx_rhs = linear_ct.cplex_num_rhs()
        cpx_sense = linear_ct.sense.cplex_code
        return cpx_lhs, cpx_sense, cpx_rhs

    def make_solution_from_vars(self, dvars):
        """ Creates an intermediate solution from a list of variables.

        :param dvars: a list of DOcplex variables.
        :return: a :class:`docplex.mp.solution.SolveSolution` object.
        """
        if dvars:
            indices = [v._index for v in dvars]
            # this calls the Cplex callback method get_values, which crashes if called with empty list
            # noinspection PyUnresolvedReferences
            var_values = super(ModelCallbackMixin, self).get_values(indices)
            # noinspection PyArgumentList
            var_value_dict = {v: val for v, val in izip(dvars, var_values)}
        else:  # pragma: no cover
            var_value_dict = {}
        return self.model.new_solution(var_value_dict)

    def make_complete_solution(self):
        """ Creates and returns an intermediate solution with all variables.

        Values are taken from the `get_values()` method of the callback

        :return: a :class:`docplex.mp.solution.SolveSolution` object.
        """
        all_vars = list(self.model.iter_variables())
        return self.make_solution_from_vars(all_vars)


class ConstraintCallbackMixin(ModelCallbackMixin):

    def __init__(self):
        ModelCallbackMixin.__init__(self)
        self._ct_vars = None
        self.cts = []
        self._vars = []

    def register_constraints(self, cts):
        self.cts.extend(cts)
        self._ct_vars = None

    def register_constraint(self, ct):
        self.register_constraints([ct])

    def register_watched_var(self, dvar):
        """ Register one variable.

        Registered variables will be part of the intermediate solutions.

        """
        self._vars.append(dvar)

    def register_watched_vars(self, dvars):
        """ Register an iterable of  variables.

        Registered variables will be part of the intermediate solutions.

        """
        self._vars.extend(dvars)

    @staticmethod
    def _collect_constraint_variables(cts):
        # collect variables as a set
        var_set = set(v for c in cts for v in c.iter_variables())
        # convert to list
        var_list = list(var_set)
        var_list.sort(key=lambda dv: dv._index)
        return var_list

    def _get_or_collect_vars(self):
        # INTERNAL
        if self._ct_vars is None:
            self._ct_vars = self._collect_constraint_variables(self.cts)
        return self._ct_vars

    @property
    def watched_vars(self):
        return self._get_or_collect_vars() + self._vars

    def make_solution_from_watched(self):
        """ Creates and returns a DOcplex solution instance from watched items.

        This method should be called when CPLEX has a new incumbent solution.
        It builds an intermediate solution from the watched variables and
        variables mentioned in the registered constraints..

        To build a soluton from all variables, use `make_complete_solution()`

        :return:
            An instance of SolveSolution.
        """
        return self.make_solution_from_vars(self.watched_vars)

    def get_cpx_unsatisfied_cts(self, cts, sol, tolerance=1e-6):
        """ returns the subset of unsatisfied constraints in a given solution.
        This is used in custom lazy constraints or user cut callbacks.

        :param cts: a list of constraints among which to look for unsatisfied
        :param sol: A solution object
        :param tolerance: amn optional numerical value used to determine
            whether a constraint is satisfied or not. Defaut is 1e-6.

        :return: a list of tuples (ct, lhs, sense, lhs) where:
            ct is an unsatisfied constraint
            lhs is the left-hand size, as expected by the cplex callback
            sense is the constraint sense, as expected by the cplex callback
            rhs is the rith-and side (a number), as expected by the cplex callback

        """
        unsatisfied = []
        for ct in cts:
            if not ct.is_satisfied(sol, tolerance):
                # use mixin API to convert to cplex lingo
                cpx_lhs, cpx_sense, cpx_rhs = self.linear_ct_to_cplex(ct)
                unsatisfied.append( (ct, cpx_lhs, cpx_sense, cpx_rhs) )
        return unsatisfied


def print_called(prompt_msg=None):
    """ A decorator function to be used on __call__() methods for derived callbacks.

    Use this decorator function to decorate __call__() methods of custom callbacks.

    Example:

    ::

        class MyCallback(ConstraintCallbackMixin, LazyConstraintCallback):

            @print_called('my custom callback called #{0}')
            def __call__(self):
                ...

        will print messages, before executing the callback code:

        >>> "my custom callback called #1"
        >>> "my custom callback called #2"

        each time the callback is called

    :param prompt_msg: A format string taking one argument (the number of calls)

    :return:
        As decorator, modifies the code of the __call_ method inplace.

    """

    def cb_decorator(func):
        prompt = prompt_msg or "* callback: {0} called: #{1}"

        def wrapper(self, *args, **kwargs):
            wrapper.count = wrapper.count + 1
            print(prompt.format(wrapper.count))
            res = func(self, *args, **kwargs)

            return res

        wrapper.count = 0
        return wrapper

    return cb_decorator




# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2015, 2016, 2017, 2018
# --------------------------------------------------------------------------

"""
Configuration of the CP Optimizer Python API

This module is the top-level handler of the configuration parameters for
the CP Optimizer Python API. It contains the default values of the different
configuration parameters.

It should NOT be changed directly.
The preferable way is to add at least one of the following files that contain the changes
to be performed:

 * *cpo_config.py*, a local set of changes on these parameters,
 * *cpo_config_<hostname>.py*, a hostname dependent set of changes.
 * *docloud_config.py* (for DOcloud url and key, file shared with docplex.mp package).

Final set of parameters is obtained by reading first this module, and then those
listed above.
These modules should be visible from the *PYTHONPATH* and are loaded in
this order to overwrite default values.

This module also defines two global variables:

 * *LOCAL_CONTEXT*, that contains the configuration appropriate to solve a model with a local
   installation of the CPO solver.
   This configuration is available for solver with version number greater or equal to 12.7.0.

   This context is the context by default, referenced by the global variable 'context'.

 * *DOCLOUD_CONTEXT*, that contains the configuration necessary to solve a model on DOcloud.

The method :meth:`set_default` allows to set the default configuration to one that is predefined,
or another that has been totally customized.

If called as main, this module prints the actual configuration on standard output, including
all customizations made using the mechanism described above.

Following sections describe the most important parameters that can be easily modified to customize
the behavior of the Python API.
All available parameters are available by consulting the source code of this module.

General parameters
------------------

*context.log_output = sys.stdout*

    This parameter contains the default log stream.
    By default it is set to the standard output.
    A value of *None* can be used to disable all logs.

*context.verbose = 0*

    This parameter controls the verbosity level of the log, between 0 and 9, if *log_output* is not None.
    The default value of 0 means no log.

*context.model.add_source_location = True*

    This parameter indicates that when the model is transformed into CPO format, additional information is added
    to correlate expressions with the Python file and line where it has been generated.
    If any error is raised by the solver during the solve, this information is provided in the
    error description, which allows for easier debugging.

*context.model.length_for_alias = None*

    This parameter allows to associate a shorter alias to variables whose name is longer than the given length.
    In the CPO representation of the model, variable is declared with its original name and an alias is created
    to use it with a shorter name in model expressions, allowing to reduce the size of the generated CPO format.

    In the returned solution, variable can be still retrieved with their original names.

    By default, the value is None, which indicates to always keep original variable names.

*context.model.name_all_constraints = False*

    This parameter enables the naming of all constraints when the model is generated in CPO format.
    It is mandatory only if the *refine conflict* function is called.
    Anyway, if the *refine conflict* function is called, and if the CPO format of the model has already been generated,
    it is generated again with this option set in order to allow proper completion of the request.
    Setting it to *True* is preferable only if *refine conflict* function is called on a big model.

*context.model.dump_directory = None*

    This parameter gives the name of a directory where the CPO files that are generated for solving models are stored
    for logging purpose.

    If not None, the directory is created and generated models are stored in files named `<model_name>.cpo`.

*context.model.cache.size = 10000*

    This parameter gives the maximum capacity of the internal cache used to speed-up conversion of Python expressions
    into CPO expressions.

*context.model.cache.active = True*

    This parameter allows to enable or disable the expression cache mechanism.
    Value os a boolean (True or False). Default value is True.

*context.params.\**

    The parameter `context.params` is an instance of the class
    :class:`~docplex.cp.parameters.CpoParameters` (in :doc:`parameters.py</docplex.cp.parameters.py>`)
    which describes all of the public solver parameters as properties.


Configuration of the model solving
----------------------------------

*context.solver.trace_cpo = False*

    This parameter indicates to trace the CPO model that is generated before submitting it for solving.
    The model is printed on the `context.log_output stream`, if given.

*context.solver.trace_log = False*

    This parameter indicates to trace the log generated by the solver when solving the CPO model.
    The log is printed on the `context.log_output stream`, if given.

    The default value of this parameter is False for a solve on the cloud, and True for a local solve.

*context.solver.enable_undocumented_params = False*

    This parameter allows to enable the possibility to set solving parameters that are not in the public parameters
    detailed in the class
    :class:`~docplex.cp.parameters.CpoParameters` (in :doc:`parameters.py</docplex.cp.parameters.py>`).

*context.solver.add_log_to_solution = True*

    This parameter indicates to add the solver log content to the solution object.
    By default, this parameter is True but it can be set to False if the log is very big or of no interest.

*context.solver.add_conflict_as_cpo = True*

    This parameter indicates to include the conflict in CPO format in the conflict refiner result
    By default, this parameter is True.

*context.solver.agent = 'local'*

    This parameter specifies the name of the solver agent that is used to solve the model.
    The value of this parameter is the name of a child context of `context.solver`, which contains necessary attributes
    that allow to create and run the required agent.

    There are two different agents described in the default configuration file:

       * `local`, the default agent, allowing to solve models locally using the CP Optimizer Interactive coming with
         versions of COS greater or equal to 12.7.0.
       * `docloud`, the agent for solving a CPO model using the DOcplexcloud service.

    If the CP Optimizer Interactive program *cpoptimizer(.exe)* is **NOT** detected in the system path,
    the default solver agent is automatically set to *docloud* instead of *local*.

*context.solver.log_prefix = "[Solver] "*

    Prefix that is added to every message that is logged by the solver component.


Configuration of the `local` solving agent
------------------------------------------

*context.solver.local.execfile*

    Name or full path of the CP Optimizer Interactive executable file.
    By default, it is set to *cpoptimizer(.exe)*, which supposes that the program is visible from the system path.


Configuration of the `docloud` solving agent
--------------------------------------------

*context.solver.docloud.url = "https://api-oaas.docloud.ibmcloud.com/job_manager/rest/v1/"*

    This parameter is used to specify the URL of the *DOcplexcloud* service.

*context.solver.docloud.key = "'Set your key in docloud_config.py'"*

    This parameter contains the personal key for authorizing access to the *DOcplexcloud* service.
    Access credentials (base URL and access key) can be retrieved after registration from `<http://developer.ibm.com/docloud/docs/api-key/>`_.

*context.solver.docloud.verify_ssl = True*

    This parameter allows to enable/disable the verification of SSL certificates.

*context.solver.docloud.proxies = None*

    This parameter allows to optionally define proxies to be used in the connection with *DOcplexcloud*.
    It is a Python dictionary protocol_name / endpoint, as described in http://docs.python-requests.org/en/master/user/advanced/#proxies.

*context.solver.docloud.request_timeout = 30*

    This parameter contains the maximum time, in seconds, that a response is waited for after a unitary request to *DOcplexcloud* server.

*context.solver.docloud.result_wait_extra_time = 60*

    This parameter is a time in seconds added to the expected solve time to compute the total result waiting timeout.

*context.solver.docloud.clean_job_after_solve = True*

    This parameter indicates whether the job is automatically cleaned after the model is solved.
    If not set to True, the model stays on the *DOcplexcloud* server and is visible from its *DropSolve* interface.
    Note that the server may block future solving requests if there are too many jobs waiting.

*context.solver.docloud.polling = Context(min=1, max=3, incr=0.2)*

    This parameter describes how the Python client polls the result of the solve on *DOcplexcloud*.
    Polling delay is inside an interval [min, max], starting by min, growing to max with the given increment.


Configuration for best performances
-----------------------------------

To configure the CP Python API for best performances, the following configuration settings may be used.
Obviously, this performance is won at the cost of the loss of some features that may be useful in other cases.
::

    context.verbose = 0
    context.model.add_source_location = False
    context.model.length_for_alias = 10
    context.model.name_all_constraints = False
    context.model.dump_directory = None
    context.solver.trace_cpo = False
    context.solver.trace_log = False
    context.solver.add_log_to_solution = False


Detailed description
--------------------
"""

from docplex.cp.utils import *
from docplex.cp.parameters import CpoParameters, ALL_PARAMETER_NAMES

import sys, socket, os, traceback

# Check if running in a worker environment
try:
    import docplex.util.environment as runenv
    IS_IN_WORKER = isinstance(runenv.get_environment(), runenv.WorkerEnvironment)
except:
    IS_IN_WORKER = False

# CP Optimizer Interactive executable name
CPO_EXEC_INTERACTIVE = "cpoptimizer" + (".exe" if IS_WINDOWS else "")

# CP Optimizer Interactive executable name
CPO_LIBRARY = "lib_cpo_solver_12100" + (".dll" if IS_WINDOWS else ".so")

# Determine path extension to search executables
python_home = os.path.dirname(os.path.abspath(sys.executable))
if IS_WINDOWS:
    PATH_EXTENSION = [os.path.join(python_home, "Scripts")]
    appdata = os.environ.get('APPDATA')
    if appdata is not None:
        PATH_EXTENSION.append(os.path.join(appdata, os.path.join('Python', 'Scripts')))
else:
    PATH_EXTENSION = ["~/.local/bin", os.path.join(python_home, "bin")]



##############################################################################
## Define default context for DOcloud solving
##############################################################################

#-----------------------------------------------------------------------------
# Global context

# Create default context infrastructure
context = Context()

# Default log output
context.log_output = sys.stdout

# Default log verbosity
context.verbose = 0

# Visu enable indicator (internal, can be disabled for testing purpose)
context.visu_enabled = True

# Indicator to log catched exceptions
context.log_exceptions = False


#-----------------------------------------------------------------------------
# Modeling context

context.model = Context()

# Indicate to add source location in model
context.model.add_source_location = True

# Minimal variable name length that trigger use of shorter alias. None for no alias.
context.model.length_for_alias = 15

# Automatically add a name to every top-level constraint
context.model.name_all_constraints = False

# Model format generation version that is used by default if no format version is given in the model.
# If None, latest format is used without specifying it explicitly.
context.model.version = None

# Name of the directory where store copy of the generated CPO files. None for no dump.
context.model.dump_directory = None

# Flag to generate short model output (internal)
context.model.short_output = False

# Expression cache
context.model.cache = Context()
context.model.cache.size = 10000
context.model.cache.active = True


#-----------------------------------------------------------------------------
# Parsing context

context.parser = Context()

# Indicate to FZN parser to reduce model when possible
context.parser.fzn_reduce = False


#-----------------------------------------------------------------------------
# Solving parameters

context.params = CpoParameters()

# Default time limit in seconds (None for no limit)
context.params.TimeLimit = None

# Workers count (None for number of cores)
context.params.Workers = None


#-----------------------------------------------------------------------------
# Solving context

context.solver = Context()

# Indicate to trace CPO model before solving
context.solver.trace_cpo = False

# Indicate to trace solver log on log_output.
context.solver.trace_log = False

# Enable undocumented parameters
context.solver.enable_undocumented_params = False

# Max number of threads allowed for model solving
context.solver.max_threads = None

# Indicate to add solver log to the solution
context.solver.add_log_to_solution = True

# Indicate to add the conflict in CPO format to conflict refiner result
context.solver.add_conflict_as_cpo = True

# Indicate to replace simple solve by a start/next loop
context.solver.solve_with_start_next = False

# Log prefix
context.solver.log_prefix = "[Solver] "

# Name of the agent to be used for solving. Value is name of one of this context child context (i.e. 'local' or 'docloud').
context.solver.agent = 'local'

# Auto-publish parameters
context.solver.auto_publish = Context()

# Indicate to auto-publish solve details in environment
context.solver.auto_publish.solve_details = True

# Indicate to auto-publish results in environment
context.solver.auto_publish.result_output = "solution.json"

# Indicate to auto-publish kpis in environment
context.solver.auto_publish.kpis_output = "kpis.csv"

# For KPIs output, name of the kpi name column
context.solver.auto_publish.kpis_output_field_name = "Name"

# For KPIs output, name of the kpi value column
context.solver.auto_publish.kpis_output_field_value = "Value"

# Indicate to auto-publish conflicts in environment
context.solver.auto_publish.conflicts_output = "conflicts.csv"

# Indicate to enable auto-publish also with local environment
context.solver.auto_publish.local_publish = False

# Default solver listeners
context.solver.listeners = ["docplex.cp.solver.environment_client.EnvSolverListener"]


#-----------------------------------------------------------------------------
# Local solving using CP Interactive executable

context.solver.local = Context()

# Python class implementing the agent
context.solver.local.class_name = "docplex.cp.solver.solver_local.CpoSolverLocal"

# Name or path of the CP Optimizer Interactive program
context.solver.local.execfile = CPO_EXEC_INTERACTIVE

# Parameters of the exec file (mandatory, do not change)
context.solver.local.parameters = ['-angel']

# Agent log prefix
context.solver.local.log_prefix = "[Local] "

# Set exec file with full path if found in the path
cpxfile = search_file_in_path(context.solver.local.execfile, PATH_EXTENSION)
if cpxfile:
    context.solver.local.execfile = cpxfile


#-----------------------------------------------------------------------------
# Local solving with CPO library (internal)

context.solver.lib = Context()

# Python class implementing the agent
context.solver.lib.class_name = "docplex.cp.solver.solver_lib.CpoSolverLib"

# Name or path of the CPO library
context.solver.lib.libfile = CPO_LIBRARY

# Agent log prefix
context.solver.lib.log_prefix = "[PyLib] "

# Check if library has been installed with specific package
try:
    from docplex_cpo_solver import get_library_path
    libfile = get_library_path()
    # Force solver to use lib by default
    context.solver.agent = 'lib'
except:
    libfile = None

# If no special lib is given, search it in the path
if libfile is None:
    libfile = search_file_in_path(context.solver.lib.libfile, PATH_EXTENSION)

# Set library file with full path if it has been found
if libfile:
        context.solver.lib.libfile = libfile


#-----------------------------------------------------------------------------
# DoCloud solving agent context

context.solver.docloud = Context()

# Python class implementing the agent
context.solver.docloud.class_name = "docplex.cp.solver.solver_docloud.CpoSolverDocloud"

# Url of the DOCloud service
context.solver.docloud.url = "https://api-oaas.docloud.ibmcloud.com/job_manager/rest/v1/"

# Authentication key.
context.solver.docloud.key = "'Set your key in docloud_config.py''"

# Secret key.
context.solver.docloud.secret = None

# Indicate to verify SSL certificates
context.solver.docloud.verify_ssl = True

# Proxies (map protocol_name/endpoint, as described in http://docs.python-requests.org/en/master/user/advanced/#proxies)
context.solver.docloud.proxies = None

# Default unitary request timeout in seconds
context.solver.docloud.request_timeout = 30

# Time added to expected solve time to compute the total result waiting timeout
context.solver.docloud.result_wait_extra_time = 60

# Clean job after solve indicator
context.solver.docloud.clean_job_after_solve = True

# Add 'Connection close' in all headers
context.solver.docloud.always_close_connection = False

# Log prefix
context.solver.docloud.log_prefix = "[DOcloud] "

# Polling delay (min, max and increment)
context.solver.docloud.polling = Context(min=1, max=3, incr=0.2)


#-----------------------------------------------------------------------------
# Create

# Create 2 contexts for local and docloud
LOCAL_CONTEXT = context
DOCLOUD_CONTEXT = context.clone()

# Set local context specific attributes
LOCAL_CONTEXT.solver.trace_log = not IS_IN_NOTEBOOK
LOCAL_CONTEXT.model.length_for_alias = None

# Set DOcloud context specific attributes
DOCLOUD_CONTEXT.solver.agent = 'docloud'
DOCLOUD_CONTEXT.solver.trace_log = False
DOCLOUD_CONTEXT.model.length_for_alias = 15


#-----------------------------------------------------------------------------
# Apply special changes if running in a worker

if IS_IN_WORKER:
    context.solver.max_threads = runenv.get_environment().get_available_core_count()


##############################################################################
## Public functions
##############################################################################

def get_default():
    """ Get the default context

    Default context is also accessible with the global variable 'context' in this module.

    Returns:
        Current default context
    """
    return context

def set_default(ctx):
    """ Set the default context.

    Default context becomes accessible in the global variable 'context' in this module.

    Args:
        ctx: New default context
    """
    if ctx is None:
        ctx = Context()
    else:
        assert isinstance(ctx, Context), "Context object must be of class Context"
    sys.modules[__name__].context = ctx


# Attribute values denoting a default value
DEFAULT_VALUES = ("ENTER YOUR KEY HERE", "ENTER YOUR URL HERE", "default")

def _is_defined(arg, kwargs):
    return (arg in kwargs) and kwargs[arg] and (kwargs[arg] not in DEFAULT_VALUES)

def _get_effective_context(**kwargs):
    """ Build a effective context from a variable list of arguments that may specify changes to default.

    Args:
        context (optional):   Source context, if not default.
        params (optional):    Solving parameters (CpoParameters) that overwrite those in the solving context
        (others) (optional):  All other context parameters that can be changed.
    Returns:
        Updated (cloned) context
    """
    # If 'url' and 'key' are defined, force agent to be docloud
    if ('agent' not in kwargs) and not IS_IN_WORKER:
        url = kwargs.get('url')
        key = kwargs.get('key')
        if url and key and is_string(url) and is_string(key) and url.startswith('http'):
            kwargs['agent'] = 'docloud'

    # Determine source context
    ctx = kwargs.get('context')
    if (ctx is None) or (ctx in DEFAULT_VALUES):
        ctx = context
    ctx = ctx.clone()
    # print("\n*** Source context");
    # ctx.write()

    # First set parameters if given
    prms = kwargs.get('params')
    if prms is not None:
        ctx.params.add(prms)

    # Process other changes
    rplist = []  # List of replacements to be done in solving parameters
    for k, v in kwargs.items():
        if (k != 'context') and (k != 'params') and (v not in DEFAULT_VALUES):
            rp = ctx.search_and_replace_attribute(k, v)
            # If not found, set in solving parameters
            if rp is None:
                rplist.append((k, v))

     # Replace or set remaining fields in parameters
    if rplist:
        params = ctx.params
        chkparams = not ctx.solver.enable_undocumented_params
        if isinstance(params, CpoParameters):
            for k, v in rplist:
                if chkparams and not k in ALL_PARAMETER_NAMES:
                    raise CpoException("CPO solver does not accept a parameter named '{}'".format(k))
                setattr(params, k, v)

    # Return
    return ctx


##############################################################################
## Overload this configuration with other customized configuration python files
##############################################################################

def _eval_file(file):
    """ If exists, evaluate the content of a python module in this module.
    Args:
        file: Python file to evaluate
    """
    for f in filter(os.path.isfile, [dir + "/" + file for dir in sys.path]):
        try:
            global context
            l = {'context': context, '__file__': os.path.abspath(f) }
            exec(open(f).read(), globals(), l)
            # Restore context in case its value has been modified.
            context = l['context']
        except Exception as e:
            if context.log_exceptions:
                traceback.print_exc()
            raise Exception("Error while loading config file {}: {}".format(f, str(e)))


# Initialize default list of files to load
FILE_LIST = ("cpo_config.py",
             "cpo_config_" + socket.gethostname() + ".py",
             "docloud_config.py")

# Load all config changes
for f in FILE_LIST:
    _eval_file(f)


##############################################################################
## Print configuration when called as main
##############################################################################

if __name__ == "__main__":
    context.write()



# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2017
# --------------------------------------------------------------------------

try:
    from sklearn.base import TransformerMixin, BaseEstimator
except ImportError:
    # if sklearn is not available, just use some mock classes
    class TransformerMixin(object):
        pass
    class BaseEstimator(object):
        pass
    
from docplex.mp.constants import ObjectiveSense
from docplex.mp.sktrans.modeler import make_modeler
from docplex.mp.utils import *

try:
    import numpy as np
except ImportError:
    np = None

class CplexTransformerBase(BaseEstimator, TransformerMixin):
    """ Root class for CPLEX transformers
    """

    def __init__(self, sense="min", modeler="cplex"):
        self.objsense = ObjectiveSense.parse(sense)  # fail if error
        self.modeler = make_modeler(modeler)

    def fit(self, *_):
        return self

    def transform(self, X, y=None, **params):
        """ Main method to solve Linear Programming problemss.

        :param X: the matrix describing the  constraints of the problem.
            Accepts numpy matrices, pandas dataframes, or sciPy sparse matrices
        :param y: an optional sequence of scalars descrining the cost vector
        :param params: optional keyword arguments to pass additional parameters.

        :return: a pandas dataframe with two columns: name and value containing the values
            of the columns.
        """
        # look for upper, lower bound columns in keyword args
        var_lbs = params.get("lbs", None)
        var_ubs = params.get("ubs", None)
        var_types = params.get("types", None)
        if is_pandas_dataframe(X):
            return self._transform_from_pandas(X, y, var_lbs, var_ubs, var_types, **params)
        elif is_numpy_matrix(X):
            return self._transform_from_numpy(X, y, var_lbs, var_ubs, var_types, **params)
        elif is_scipy_sparse(X):
            return self._transform_from_sparse(X, y, var_lbs, var_ubs, var_types, **params)
        elif isinstance(X, list):
            return self._transform_from_sequence(X, y, var_lbs, var_ubs, var_types, **params)
        else:
            raise ValueError(
                'transformer expects pandas dataframe, numpy matrix or python list, {0} was passed'.format(X))

    def _transform_from_sequence(self, X, y, var_lbs, var_ubs, var_types, **params):
        raise NotImplemented

    def _transform_from_pandas(self, X, y, var_lbs, var_ubs, var_types, **params):
        raise NotImplemented

    def _transform_from_numpy(self, X, y, var_lbs, var_ubs, var_types, **params):
        raise NotImplemented

    def _transform_from_sparse(self, X, y, var_lbs, var_ubs, var_types, **params):
        raise NotImplemented

    def build_matrix_linear_model_and_solve(self, var_count, var_lbs, var_ubs, var_types, var_names,
                                            cts_mat, rhs,
                                            objsense, costs, cast_to_float,
                                            **params):
        return self.modeler.build_matrix_linear_model_and_solve(var_count, var_lbs, var_ubs,
                                                                var_types, var_names,
                                                                cts_mat, rhs,
                                                                objsense, costs,
                                                                cast_to_float, **params)

    def build_matrix_range_model_and_solve(self, var_count, var_lbs, var_ubs,
                                           var_types, var_names,
                                           cts_mat, range_mins, range_maxs,
                                           objsense, costs, cast_to_float,
                                           **params):
        return self.modeler.build_matrix_range_model_and_solve(var_count, var_lbs, var_ubs,
                                                               var_types, var_names,
                                                               cts_mat, range_mins, range_maxs,
                                                               objsense, costs,
                                                               cast_to_float,
                                                               **params)

    def build_sparse_linear_model_and_solve(self, nb_vars, var_lbs, var_ubs, var_types, var_names,
                                            nb_rows, cts_sparse_coefs,
                                            objsense, costs,
                                            **params):
        return self.modeler.build_sparse_linear_model_and_solve(nb_vars, var_lbs, var_ubs,
                                                                var_types, var_names,
                                                                nb_rows, cts_sparse_coefs,
                                                                objsense, costs,
                                                                **params)


class CplexTransformer(CplexTransformerBase):
    """ A Scikit-learn transformer class to solve linear problems.

    This transformer class solves LP problems of
    type::

            Ax <= B

    """

    def __init__(self, sense="min", modeler="cplex"):
        """
        Creates an instance of LPTransformer to solve linear problems.

        :param sense: defines the objective sense. Accepts 'min" or "max" (not case-sensitive),
            or an instance of docplex.mp.ObjectiveSense

        Note:
            The matrix X is supposed to have shape (M,N+1) where M is the number of rows
            and N the number of variables. The last column contains the right hand sides of the problem
            (the B in Ax <= B)
            The optional vector Y contains the N cost coefficients for each column variables.

        Example:
            Passing X = [[1,2,3], [4,5,6]], Y= [11,12,13] means solving the linear problem:

                minimize 11x + 12y + 13z
                s.t.
                        1x + 2y <= 3
                        4x + 5y <= 6
        """
        super(CplexTransformer, self).__init__(sense, modeler)

    def _transform_from_pandas(self, X, y, var_lbs, var_ubs, var_types, **params):
        assert is_pandas_dataframe(X)

        X_new = X.copy()
        # save min, max per nutrients in lists, drop them
        rhs = X["rhs"].tolist()
        X_new.drop(labels=["rhs"], inplace=True, axis=1)
        _, x_cols = X.shape
        nb_vars = x_cols - 1
        return self.build_matrix_linear_model_and_solve(nb_vars, var_lbs, var_ubs, var_types,
                                                        var_names=X_new.columns,
                                                        cts_mat=X_new, rhs=rhs,
                                                        objsense=self.objsense, costs=y,
                                                        **params)

    def _transform_from_sequence(self, X, y, var_lbs, var_ubs, var_types, **params):
        # matrix is a list of lists nr x [nc+1]
        # last two columns are lbs, ubs in that order
        assert X
        xc = len(X[0])
        for r in X:
            assert xc == len(r)

        colnames = params.pop("colnames", None)

        nb_vars = xc - 1
        X_cts = [r[:-1] for r in X]
        rhs = [r[-1] for r in X]
        # no cast to float except if requested
        cast_to_float = params.pop("cast_to_float", False)
        return self.build_matrix_linear_model_and_solve(nb_vars, var_lbs, var_ubs, var_types,
                                                        colnames,
                                                        X_cts, rhs,
                                                        objsense=self.objsense, costs=y,
                                                        cast_to_float=cast_to_float,
                                                        **params)

    def _transform_from_numpy(self, X, y, var_lbs, var_ubs, var_types, **params):
        # matrix is nrows x (ncols + 2)
        # last two columns are lbs, ubs in that order
        assert is_numpy_matrix(X)

        colnames = params.pop("colnames", None)
        mshape = X.shape
        nr, nc = mshape

        assert nc >= 2
        nb_vars = nc - 1
        X_cts = X[:, :-1]
        rhs = X[:, -1].A1
        # to cast or not to cast?
        cast_to_float = (X.dtype == np.int64) or params.pop("cast-to_float", False)
        return self.build_matrix_linear_model_and_solve(nb_vars, var_lbs, var_ubs, var_types,colnames,
                                                        X_cts, rhs,
                                                        objsense=self.objsense, costs=y,
                                                        cast_to_float=cast_to_float, **params)

    def _transform_from_sparse(self, X, y, var_lbs, var_ubs, var_types, **params):
        assert is_scipy_sparse(X)

        colnames = params.pop("colnames", None)
        mshape = X.shape
        nr, nc = mshape
        nb_vars = nc - 1
        #  convert to coo before iterate()
        x_coo = X.tocoo()
        cts_sparse_coefs = izip(x_coo.data, x_coo.row, x_coo.col)
        return self.build_sparse_linear_model_and_solve(nb_vars, var_lbs, var_ubs, var_types,colnames,
                                                        nr, cts_sparse_coefs,
                                                        objsense=self.objsense, costs=y, **params)


class CplexRangeTransformer(CplexTransformerBase):
    def __init__(self, sense="min", modeler="cplex"):
        """
        Creates an instance of LPRangeTransformer to solve range-based linear problems.

        :param sense: defines the objective sense. Accepts 'min" or "max" (not case-sensitive),
            or an instance of docplex.mp.ObjectiveSense

        Note:
            The matrix X is supposed to have shape (M,N+2) where M is the number of rows
            and N the number of variables.
            The last two columns are assumed to contain the minimum (resp.maximum) values for the
            row ranges, that m and M in:
                    m <= Ax <= M

            The optional vector Y contains the N cost coefficients for each column variables.

        Example:
            Passing X = [[1,2,3,30], [4,5,6,60]], Y= [11,12,13] means solving the linear problem:

                minimize 11x + 12y + 13z
                s.t.
                        3 <= 1x + 2y <= 30
                        6 <= 4x + 5y <= 60
        """
        super(CplexRangeTransformer, self).__init__(sense, modeler)

    def _transform_from_sequence(self, X, y, var_lbs, var_ubs, **params):
        # matrix is a list of lists nr x [nc+1]
        # last two columns are lbs, ubs in that order
        assert X
        xc = len(X[0])
        assert xc >= 3  # one var plus min max
        for r in X:
            assert xc == len(r)

        colnames = params.get("colnames", None)

        nb_vars = xc - 2
        X_cts = [r[:-2] for r in X]
        row_maxs = [r[-1] for r in X]
        row_mins = [r[-2] for r in X]
        return self.build_matrix_range_model_and_solve(nb_vars, var_lbs, var_ubs, colnames,
                                                       cts_mat=X_cts,
                                                       range_mins=row_mins, range_maxs=row_maxs,
                                                       objsense=self.objsense, costs=y,
                                                       **params)

    def _transform_from_pandas(self, X, y, var_lbs, var_ubs, var_types, **params):
        assert is_pandas_dataframe(X)

        x_rows, x_cols = X.shape
        X_new = X.copy()
        # extract columns with name 'min' and 'max' as series then drop
        row_mins = X["min"].tolist()
        row_maxs = X["max"].tolist()
        X_new.drop(labels=["min", "max"], inplace=True, axis=1)
        nb_vars = x_cols - 2
        varnames = X_new.columns.values.tolist()
        return self.build_matrix_range_model_and_solve(nb_vars, var_lbs, var_ubs, var_types,
                                                       varnames,
                                                       cts_mat=X_new,
                                                       range_mins=row_mins, range_maxs=row_maxs,
                                                       objsense=self.objsense, costs=y,
                                                       cast_to_float=True,
                                                       **params)

    def _transform_from_numpy(self, X, y, var_lbs, var_ubs, var_types, **params):
        # matrix is nrows x (ncols + 2)
        # last two columns are lbs, ubs in that order
        assert is_numpy_matrix(X)

        colnames = params.pop("colnames", None)
        mshape = X.shape
        xr, xc = mshape
        assert xc >= 3
        nb_vars = xc - 2
        X_cts = X[:, :-2]
        row_mins = X[:, -2]
        row_maxs = X[:, -1]
        cast_to_float = (X.dtype == np.int64) or params.pop("cast_to_float", False)
        return self.build_matrix_range_model_and_solve(nb_vars, var_lbs, var_ubs,
                                                       var_types, colnames,
                                                       cts_mat=X_cts,
                                                       range_mins=row_mins, range_maxs=row_maxs,
                                                       objsense=self.objsense, costs=y,
                                                       cast_to_float=cast_to_float,
                                                       **params)



# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2015, 2016
# --------------------------------------------------------------------------

# gendoc: ignore
from docplex.mp.mfactory import _AbstractModelFactory

from docplex.mp.constants import ComparisonType, UpdateEvent
from docplex.mp.utils import is_number
from docplex.mp.xcounter import update_dict_from_item_value
from docplex.mp.quad import QuadExpr, VarPair
from docplex.mp.operand import LinearOperand
from docplex.mp.linear import Var, MonomialExpr, ZeroExpr, AbstractLinearExpr
from docplex.mp.constr import QuadraticConstraint


class IQuadFactory(_AbstractModelFactory):
    # INTERNAL

    def new_var_product(self, var, factor):
        raise NotImplementedError  # pragma: no cover

    def new_monomial_product(self, monexpr, factor):
        raise NotImplementedError  # pragma: no cover

    def new_linexpr_product(self, linexpr, factor):
        raise NotImplementedError  # pragma: no cover

    def new_var_square(self, var):
        raise NotImplementedError  # pragma: no cover


class QuadFactory(IQuadFactory):
    def __init__(self, model, engine):
        _AbstractModelFactory.__init__(self, model)
        self._model = model
        self._engine = engine
        self._lfactory = model._lfactory
        self._quad_count = 0

    def new_zero_expr(self):
        return ZeroExpr(self._model)

    def _unexpected_product_error(self, factor1, factor2):
        # INTERNAL
        self._model.fatal("cannot multiply {0!s} by {1!s}", factor1, factor2)

    def new_quad(self, quads=None, linexpr=None, name=None, safe=False):
        self._quad_count += 1
        return QuadExpr(self._model, quads=quads, linexpr=linexpr, name=name, safe=safe)

    def new_linear_expr(self, e=0, cst=0):
        return self._lfactory.linear_expr(e, cst)

    def new_var_square(self, var):
        return self.new_quad(quads=(var, var, 1), safe=True)

    def new_var_product(self, var, other):
        # computes and returns the product var * other
        if isinstance(other, Var):
            return self.new_quad(quads=(var, other, 1), safe=True)
        elif isinstance(other, MonomialExpr):
            mnm_dvar = other._dvar
            mnm_coef = other.coef
            return self.new_quad(quads=(var, mnm_dvar, mnm_coef), safe=True)
        elif isinstance(other, ZeroExpr):
            return other
        elif isinstance(other, LinearOperand):
            linear_op = other
            quad_args = [(var, dv, k) for dv, k in linear_op.iter_terms()]
            linexpr_k = linear_op.get_constant()
            quad_linexpr = linexpr_k * var if linexpr_k else None
            return self.new_quad(quad_args, quad_linexpr, safe=True)

        # elif isinstance(other, FunctionalExpr):
        #     return self.new_quad(quads=(var, other.as_var, 1), safe=True)

        else:
            self._unexpected_product_error(var, other)

    def new_monomial_product(self, mnm, other):
        mnmk = mnm.coef
        if not mnmk:  # pragma: no cover
            return self.new_zero_expr()
        else:
            var_quad = self.new_var_product(mnm.var, other)
            var_quad._scale(mnmk)
            return var_quad

    def new_linexpr_product(self, linexpr, other):
        if isinstance(other, Var):
            return self.new_var_product(other, linexpr)

        elif isinstance(other, MonomialExpr):
            return self.new_monomial_product(other, linexpr)

        elif isinstance(other, LinearOperand):
            cst1 = linexpr.constant
            cst2 = other.get_constant()

            fcc = self.term_dict_type()
            for lv1, lk1 in linexpr.iter_terms():
                for lv2, lk2 in other.iter_terms():
                    update_dict_from_item_value(fcc, VarPair(lv1, lv2), lk1 * lk2)
            # this is quad
            qlinexpr = self.new_linear_expr()
            # add cst2 * linexp1
            qlinexpr._add_expr_scaled(expr=linexpr, factor=cst2)
            # add cst1 * linexpr2
            qlinexpr._add_expr_scaled(expr=other, factor=cst1)

            # and that's it
            # fix the constant
            qlinexpr.constant = cst1 * cst2
            quad = QuadExpr(self._model, quads=fcc, linexpr=qlinexpr, safe=True)
            return quad

        else:
            self._unexpected_product_error(linexpr, other)

    def new_xconstraint(self, lhs, rhs, comparaison_type):
        if self._quad_count and isinstance(rhs, QuadExpr):
            return self._new_qconstraint(lhs, comparaison_type, rhs)
        else:
            return self._lfactory._new_binary_constraint(lhs, comparaison_type, rhs)

    def new_le_constraint(self, e, rhs, ctname=None):
        return self._new_qconstraint(e, ComparisonType.LE, rhs, name=ctname)

    def new_eq_constraint(self, e, rhs, ctname=None):
        return self._new_qconstraint(e, ComparisonType.EQ, rhs, name=ctname)

    def new_ge_constraint(self, e, rhs, ctname=None):
        return self._new_qconstraint(e, ComparisonType.GE, rhs, name=ctname)

    def _new_qconstraint(self, lhs, ct_sense, rhs, name=None):
        # noinspection PyPep8
        left_expr = self._to_expr(lhs, context="QuadraticConstraint.left_expr")
        right_expr = self._to_expr(rhs, context="QuadraticConstraint.right_expr")
        self._model._checker.typecheck_two_in_model(self._model, left_expr, right_expr, "new_binary_constraint")
        ct = QuadraticConstraint(self._model, left_expr, ct_sense, right_expr, name)
        left_expr.notify_used(ct)
        right_expr.notify_used(ct)
        return ct

    def _to_expr(self, e, context=None):
        # INTERNAL
        if isinstance(e, (AbstractLinearExpr, QuadExpr)):
            return e
        elif is_number(e):
            return self._lfactory.constant_expr(cst=e)
        else:
            try:
                return e.to_linear_expr()

            except AttributeError:  # pragma: no cover
                self._model.fatal("cannot convert to expression: {0!r}", e)

    def set_quadratic_constraint_expr_from_pos(self, qct, pos, new_expr, update_subscribers=True):
        old_expr = qct.get_expr_from_pos(pos)
        new_operand = self._to_expr(e=new_expr)
        exprs = [qct._left_expr, qct._right_expr]
        exprs[pos] = new_operand
        self._engine.update_constraint(qct, UpdateEvent.LinearConstraintGlobal, *exprs)
        # discard old_expr
        qct.set_expr_from_pos(pos, new_expr=new_operand)
        old_expr.notify_unsubscribed(qct)
        if update_subscribers:
            # -- update  subscribers
            old_expr.notify_unsubscribed(qct)
            new_operand.notify_used(qct)

    def set_quadratic_constraint_sense(self, qct, arg_newsense):
        new_sense = ComparisonType.parse(arg_newsense)
        if new_sense != qct.sense:
            self._engine.update_constraint(qct, UpdateEvent.ConstraintSense, new_sense)
            qct._internal_set_sense(new_sense)

    def update_quadratic_constraint(self, qct, expr, event):
        self._engine.update_constraint(qct, event, expr)

